""" Structure file for ThermoFisher Genesys 30 Adapter """
from __future__ import annotations

from datetime import datetime
from pathlib import Path

import pandas as pd

from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    ContainerType,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.thermo_fisher_genesys30 import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(header: SeriesData, file_path: str) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        device_type=constants.DEVICE_TYPE,
        device_identifier=NOT_APPLICABLE,
        model_number=constants.MODEL_NUMBER,
        software_name=constants.GENESYS_SOFTWARE,
        detection_type=header.get(str, "Mode"),
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        brand_name=constants.BRAND_NAME,
        container_type=ContainerType.tube,
    )


def create_measurement_groups(
    header: SeriesData, data: pd.DataFrame
) -> list[MeasurementGroup]:
    experiment_name = header.get(str, "Scan")
    experiment_type, measurement_time = _get_experiment_type_and_time(experiment_name)
    return [
        MeasurementGroup(
            measurement_time=measurement_time,
            experiment_type=experiment_type,
            measurements=[
                Measurement(
                    type_=MeasurementType.ULTRAVIOLET_ABSORBANCE_SPECTRUM,
                    identifier=random_uuid_str(),
                    sample_identifier=NOT_APPLICABLE,
                    data_cube=DataCube(
                        label="absorption spectrum",
                        structure_dimensions=[
                            DataCubeComponent(
                                type_=FieldComponentDatatype.double,
                                concept="wavelength",
                                unit="nm",
                            )
                        ],
                        structure_measures=[
                            DataCubeComponent(
                                type_=FieldComponentDatatype.double,
                                concept="absorbance",
                                unit="mAU",
                            )
                        ],
                        dimensions=[data["wavelength(nm)"].tolist()],
                        measures=[data["ABS"].astype(float).tolist()],
                    ),
                    device_control_custom_info={
                        "operating minimum": {
                            "value": header.get(float, "Lower"),
                            "unit": "nm",
                        },
                        "operating maximum": {
                            "value": header.get(float, "Upper"),
                            "unit": "nm",
                        },
                    },
                ),
            ],
        ),
    ]


def _get_experiment_type_and_time(experiment_name: str | None) -> tuple[str, str]:
    """
    Gets data and time of measurement from the experiment name
    :param experiment_name: The name of the experiment conducted
    :return: Experiment type and measurement time
    """
    if experiment_name:
        experiment_type, date, time = experiment_name.split("_", 2)
        time = time.split(".")[0]
        timestamp = date + time

        measurement_time = (
            datetime.strptime(timestamp, "%Y%m%d%H%M%S")
            .astimezone()
            .strftime("%d-%m-%Y %I:%M %p")
        )
        return experiment_type, measurement_time
    else:
        raise AllotropeConversionError(constants.MEASUREMENT_TIME_MISSING)



from allotropy.allotrope.models.adm.liquid_handler.benchling._2024._11.liquid_handler import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.liquid_handler.benchling._2024._11.liquid_handler import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_coulter_biomek.beckman_coulter_biomek_reader import (
    BeckmanCoulterBiomekReader,
)
from allotropy.parsers.beckman_coulter_biomek.beckman_coulter_biomek_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.beckman_coulter_biomek.constants import DISPLAY_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class BeckmanCoulterBiomekParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = BeckmanCoulterBiomekReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = BeckmanCoulterBiomekReader(named_file_contents)
        return Data(
            create_metadata(reader.header, named_file_contents.original_file_path),
            create_measurement_groups(reader.data, reader.header),
        )



from io import StringIO

import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.utils.pandas import read_csv, SeriesData


class NanodropEightReader:
    SUPPORTED_EXTENSIONS = "txt,tsv"
    header: SeriesData
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader(read_to_lines(named_file_contents))

        header_data = {}
        # Header lines are expected to have a single 'key: value' pair, while table will have multiple
        # tab-separated column headers. So, detect header lines as:
        #   <anything but a tab>:<any number of tabs><anything but a tab><any whitespace>
        for line in reader.pop_while(match_pat=r"^[^\t]+:[\t]*[^\t]+[\s]*$"):
            key, value = line.split(":")
            header_data[key] = value.strip()

        header = pd.Series(header_data)
        header.index = header.index.astype(str).str.strip().str.lower()
        self.header = SeriesData(header)

        lines = reader.pop_csv_block_as_lines()
        if not lines:
            msg = "Reached end of file without finding table data."
            raise AllotropeConversionError(msg)

        self.data = read_csv(
            StringIO("\n".join(lines)),
            sep="\t",
            dtype={"Sample Name": str, "Sample ID": str},
            # Prevent pandas from rounding decimal values, at the cost of some speed.
            float_precision="round_trip",
        )
        self.data.columns = self.data.columns.str.lower()



from __future__ import annotations

from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.beckman_vi_cell_xr.constants import (
    DEFAULT_ANALYST,
    DETECTION_TYPE,
    DEVICE_TYPE,
    MODEL_NUMBER,
    SOFTWARE_NAME,
)
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_reader import ViCellData
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_measurement_group(data: SeriesData) -> MeasurementGroup:
    total_cell_count = data.get(float, "Total cells")
    total_cell_count = (
        total_cell_count if total_cell_count is None else round(total_cell_count)
    )
    viable_cell_count = data.get(float, "Viable cells")
    viable_cell_count = (
        viable_cell_count if viable_cell_count is None else round(viable_cell_count)
    )

    return MeasurementGroup(
        analyst=DEFAULT_ANALYST,
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=data[str, "Sample date"],
                sample_identifier=data[str, "Sample ID"],
                cell_type_processing_method=data.get(str, "Cell type"),
                cell_density_dilution_factor=data.get(float, "Dilution factor"),
                viability=data[float, "Viability (%)"],
                viable_cell_density=data[float, "Viable cells/ml (x10^6)"],
                total_cell_count=total_cell_count,
                total_cell_density=data.get(float, "Total cells/ml (x10^6)"),
                average_total_cell_diameter=data.get(float, "Avg. diam. (microns)"),
                viable_cell_count=viable_cell_count,
                average_total_cell_circularity=data.get(float, "Avg. circ."),
            )
        ],
    )


def create_metadata(reader_data: ViCellData, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        device_type=DEVICE_TYPE,
        detection_type=DETECTION_TYPE,
        model_number=MODEL_NUMBER,
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        equipment_serial_number=reader_data.serial_number,
        software_name=SOFTWARE_NAME,
        software_version=reader_data.version.value,
        file_name=path.name,
        unc_path=file_path,
    )



DEVICE_IDENTIFIER = "Nanodrop"
DEVICE_TYPE = "absorbance detector"
MODEL_NUBMER = "Nanodrop 8000"

# These may be reported in the results, and are stored as calculated data
ABSORBANCE_RATIOS = [
    (260, 280),
    (260, 230),
]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.allotrope.schema_mappers.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    Analyte,
    Calibration,
    Error as MapperError,
    Measurement as MapperMeasurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.luminex_xponent import constants
from allotropy.parsers.luminex_xponent.luminex_xponent_reader import (
    LuminexXponentReader,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
)


@dataclass(frozen=True)
class Header:
    model_number: str
    software_version: str
    equipment_serial_number: str
    analytical_method_identifier: str
    method_version: str
    experimental_data_identifier: str
    sample_volume_setting: float
    plate_well_count: float
    measurement_time: str
    detector_gain_setting: str
    data_system_instance_identifier: str
    minimum_assay_bead_count_setting: float
    analyst: str | None = None

    @classmethod
    def create(
        cls, header_data: pd.DataFrame, minimum_assay_bead_count_setting: float
    ) -> Header:
        info_row = SeriesData(header_data.iloc[0])
        raw_datetime = info_row[str, "BatchStartTime"]
        sample_volume = info_row[str, "SampleVolume"]

        return Header(
            model_number=cls._get_model_number(header_data),
            software_version=info_row[str, "Build"],
            equipment_serial_number=info_row[str, "SN"],
            analytical_method_identifier=info_row[str, "ProtocolName"],
            method_version=info_row[str, "ProtocolVersion"],
            experimental_data_identifier=info_row[str, "Batch"],
            sample_volume_setting=try_float(
                sample_volume.split()[0], "sample volume setting"
            ),
            plate_well_count=cls._get_plate_well_count(header_data),
            measurement_time=raw_datetime,
            detector_gain_setting=info_row[str, "ProtocolReporterGain"],
            data_system_instance_identifier=info_row[str, "ComputerName"],
            minimum_assay_bead_count_setting=minimum_assay_bead_count_setting,
            analyst=info_row.get(str, "Operator"),
        )

    @classmethod
    def _get_model_number(cls, header_data: pd.DataFrame) -> str:
        program_data = cls._try_col_from_header(header_data, "Program")

        try:
            model_number = program_data.iloc[2]
        except IndexError as e:
            msg = (
                "Unable to find model number in Program row, expected value at index 2"
            )
            raise AllotropeConversionError(msg) from e

        return str(model_number)

    @classmethod
    def _get_plate_well_count(cls, header_data: pd.DataFrame) -> float:
        protocol_plate_data = cls._try_col_from_header(header_data, "ProtocolPlate")

        try:
            plate_well_count = protocol_plate_data.iloc[3]
        except IndexError as e:
            msg = "Unable to find plate well count in ProtocolPlate row, expected value at index 3."
            raise AllotropeConversionError(msg) from e

        return try_float(str(plate_well_count), "plate well count")

    @classmethod
    def _try_col_from_header(
        cls, header_data: pd.DataFrame, key: str
    ) -> pd.Series[str]:
        if key not in header_data:
            msg = f"Unable to find {key} data in header block."
            raise AllotropeConversionError(msg)

        return header_data[key]


def create_calibration(calibration_data: SeriesData) -> Calibration:
    """Create a CalibrationItem from a calibration line.

    Each line should follow the pattern "Last <calibration_name>,<calibration_report> <calibration_time><,,,,"
    example: "Last F3DeCAL1 Calibration,Passed 05/17/2023 09:25:11,,,,,,"
    """
    if len(calibration_data.series.index) < constants.MINIMUM_CALIBRATION_LINE_COLS:
        msg = f"Expected at least two columns on the calibration line, got:\n{calibration_data.series}."
        raise AllotropeConversionError(msg)

    calibration_result = calibration_data.series.iloc[1].split(maxsplit=1)
    if len(calibration_result) != constants.EXPECTED_CALIBRATION_RESULT_LEN:
        msg = f"Invalid calibration result format, expected to split into two values, got: {calibration_result}."
        raise AllotropeConversionError(msg)

    return Calibration(
        name=calibration_data.series.iloc[0].replace("Last", "").strip(),
        report=calibration_result[0],
        time=calibration_result[1],
    )


@dataclass(frozen=True)
class Measurement:
    identifier: str
    sample_identifier: str
    location_identifier: str
    dilution_factor_setting: float
    assay_bead_count: float
    analytes: list[Analyte]
    errors: list[str] | None = None

    @classmethod
    def create(
        cls,
        median_data: SeriesData,
        count_data: pd.DataFrame,
        bead_ids_data: SeriesData,
        dilution_factor_data: pd.DataFrame,
        errors_data: pd.DataFrame,
    ) -> Measurement:
        location = str(median_data.series.name)
        if location not in dilution_factor_data.index:
            msg = f"Could not find 'Dilution Factor' data for: '{location}'."
            raise AllotropeConversionError(msg)
        if location not in count_data.index:
            msg = f"Could not find 'Count' data for: '{location}'."
            raise AllotropeConversionError(msg)

        # Keys in the median data that are not analyte data.
        metadata_keys = ["Sample", "Total Events"]

        well_location, location_id = cls._get_location_details(location)
        return Measurement(
            identifier=random_uuid_str(),
            sample_identifier=median_data[str, "Sample"],
            location_identifier=location_id,
            dilution_factor_setting=SeriesData(dilution_factor_data.loc[location])[
                float, "Dilution Factor"
            ],
            assay_bead_count=median_data[float, "Total Events"],
            analytes=[
                Analyte(
                    identifier=random_uuid_str(),
                    name=analyte,
                    assay_bead_identifier=bead_ids_data[str, analyte],
                    assay_bead_count=SeriesData(count_data.loc[location])[
                        float, analyte
                    ],
                    value=median_data[float, analyte],
                    statistic_datum_role=TStatisticDatumRole.median_role,
                )
                for analyte in [
                    key for key in median_data.series.index if key not in metadata_keys
                ]
            ],
            errors=cls._get_errors(errors_data, well_location),
        )

    @classmethod
    def _get_location_details(cls, location: str) -> tuple[str, str]:
        location_regex = r"\d+\((?P<well_location>\d+,(?P<location_id>\w+))\)"
        match = assert_not_none(
            re.search(location_regex, location),
            msg=f"Invalid location format: {location}",
        )

        return match.group("well_location"), match.group("location_id")

    @classmethod
    def _get_errors(
        cls, errors_data: pd.DataFrame, well_location: str
    ) -> list[str] | None:
        if well_location not in errors_data.index:
            return None
        return map_rows(
            errors_data.loc[[well_location]], lambda data: data[str, "Message"]
        )


@dataclass(frozen=True)
class MeasurementList:
    measurements: list[Measurement]

    @classmethod
    def create(cls, results_data: dict[str, pd.DataFrame]) -> MeasurementList:
        if missing_sections := [
            section
            for section in constants.EXPECTED_SECTIONS
            if section not in results_data
        ]:
            msg = f"Unable to parse input file, missing expected sections: {missing_sections}."
            raise AllotropeConversionError(msg)

        if "BeadID:" not in results_data["Units"].index:
            msg = (
                "Could not parse bead data from 'Units' section, missing 'BeadID:' row"
            )
            raise AllotropeConversionError()
        bead_ids_data = SeriesData(results_data["Units"].loc["BeadID:"])

        def create_measurement(data: SeriesData) -> Measurement:
            return Measurement.create(
                median_data=data,
                count_data=results_data["Count"],
                bead_ids_data=bead_ids_data,
                dilution_factor_data=results_data["Dilution Factor"],
                errors_data=results_data["Warnings/Errors"],
            )

        return MeasurementList(map_rows(results_data["Median"], create_measurement))


@dataclass(frozen=True)
class Data:
    header: Header
    calibrations: list[Calibration]
    measurement_list: MeasurementList

    @classmethod
    def create(cls, reader: LuminexXponentReader) -> Data:
        return Data(
            header=Header.create(
                reader.header_data, reader.minimum_assay_bead_count_setting
            ),
            calibrations=map_rows(reader.calibration_data, create_calibration),
            measurement_list=MeasurementList.create(reader.results_data),
        )


def create_metadata(
    header: Header, calibrations: list[Calibration], file_path: str
) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        equipment_serial_number=header.equipment_serial_number,
        model_number=header.model_number,
        calibrations=calibrations,
        data_system_instance_identifier=header.data_system_instance_identifier,
        software_name=constants.DEFAULT_SOFTWARE_NAME,
        software_version=header.software_version,
        device_type=constants.DEFAULT_DEVICE_TYPE,
    )


def create_measurement_groups(
    measurements: list[Measurement], header: Header
) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(
            analyst=header.analyst,
            analytical_method_identifier=header.analytical_method_identifier,
            method_version=header.method_version,
            experimental_data_identifier=header.experimental_data_identifier,
            container_type=constants.DEFAULT_CONTAINER_TYPE,
            plate_well_count=header.plate_well_count,
            measurements=[
                MapperMeasurement(
                    identifier=measurement.identifier,
                    measurement_time=header.measurement_time,
                    sample_identifier=measurement.sample_identifier,
                    location_identifier=measurement.location_identifier,
                    sample_volume_setting=header.sample_volume_setting,
                    dilution_factor_setting=measurement.dilution_factor_setting,
                    minimum_assay_bead_count_setting=header.minimum_assay_bead_count_setting,
                    detector_gain_setting=header.detector_gain_setting,
                    assay_bead_count=measurement.assay_bead_count,
                    analytes=measurement.analytes,
                    errors=[
                        MapperError(error=error) for error in (measurement.errors or [])
                    ],
                )
            ],
        )
        for measurement in measurements
    ]



from __future__ import annotations

from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.bmg_mars.bmg_mars_reader import BmgMarsReader
from allotropy.parsers.bmg_mars.bmg_mars_structure import (
    create_measurement_groups,
    create_metadata,
    Header,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class BmgMarsParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "BMG MARS"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "csv"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = BmgMarsReader(named_file_contents)
        header = Header.create(reader.header, reader.header_content)
        return Data(
            create_metadata(header, named_file_contents.original_file_path),
            create_measurement_groups(reader.data, header),
        )



from dataclasses import dataclass

from allotropy.allotrope.models.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    CalibrationAggregateDocument,
    CalibrationDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    MultiAnalyteProfilingAggregateDocument,
    MultiAnalyteProfilingDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class Analyte:
    identifier: str
    name: str
    value: float
    assay_bead_identifier: str
    assay_bead_count: float
    statistic_datum_role: TStatisticDatumRole | None


@dataclass(frozen=True)
class Error:
    error: str
    feature: str | None = None


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    sample_identifier: str
    location_identifier: str
    measurement_time: str

    # Settings
    assay_bead_count: float

    analytes: list[Analyte]
    errors: list[Error] | None = None

    # Optional metadata
    description: str | None = None
    sample_role_type: SampleRoleType | None = None
    well_plate_identifier: str | None = None

    # Optional settings
    sample_volume_setting: float | None = None
    dilution_factor_setting: float | None = None
    minimum_assay_bead_count_setting: float | None = None
    detector_gain_setting: str | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: float
    analyst: str | None
    analytical_method_identifier: str | None = None
    experimental_data_identifier: str | None = None
    method_version: str | None = None
    container_type: str | None = None
    experiment_type: str | None = None


@dataclass(frozen=True)
class Calibration:
    name: str
    report: str
    time: str


@dataclass(frozen=True)
class Metadata:
    file_name: str
    unc_path: str
    device_type: str
    model_number: str | None = None
    software_name: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    data_system_instance_identifier: str | None = None
    firmware_version: str | None = None
    product_manufacturer: str | None = None

    calibrations: list[Calibration] | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/multi-analyte-profiling/BENCHLING/2024/01/multi-analyte-profiling.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            multi_analyte_profiling_aggregate_document=MultiAnalyteProfilingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=data.metadata.model_number,
                    product_manufacturer=data.metadata.product_manufacturer,
                    firmware_version=data.metadata.firmware_version,
                    equipment_serial_number=data.metadata.equipment_serial_number,
                    calibration_aggregate_document=self._get_calibration_aggregate_document(
                        data.metadata.calibrations
                    ),
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                multi_analyte_profiling_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> MultiAnalyteProfilingDocumentItem:
        return MultiAnalyteProfilingDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                experiment_type=measurement_group.experiment_type,
                analytical_method_identifier=measurement_group.analytical_method_identifier,
                method_version=measurement_group.method_version,
                experimental_data_identifier=measurement_group.experimental_data_identifier,
                container_type=measurement_group.container_type,
                plate_well_count=quantity_or_none(
                    TQuantityValueNumber, measurement_group.plate_well_count
                ),
                measurement_document=[
                    self._get_measurement_document(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.measurement_time),
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        sample_volume_setting=quantity_or_none(
                            TQuantityValueMicroliter, measurement.sample_volume_setting
                        ),
                        dilution_factor_setting=quantity_or_none(
                            TQuantityValueUnitless, measurement.dilution_factor_setting
                        ),
                        detector_gain_setting=measurement.detector_gain_setting,
                        minimum_assay_bead_count_setting=quantity_or_none(
                            TQuantityValueNumber,
                            measurement.minimum_assay_bead_count_setting,
                        ),
                    )
                ]
            ),
            assay_bead_count=TQuantityValueNumber(value=measurement.assay_bead_count),
            analyte_aggregate_document=AnalyteAggregateDocument(
                analyte_document=[
                    AnalyteDocumentItem(
                        analyte_identifier=analyte.identifier,
                        analyte_name=analyte.name,
                        assay_bead_identifier=analyte.assay_bead_identifier,
                        assay_bead_count=TQuantityValueNumber(
                            value=analyte.assay_bead_count
                        ),
                        fluorescence=TQuantityValueRelativeFluorescenceUnit(
                            value=analyte.value,
                            has_statistic_datum_role=analyte.statistic_datum_role,
                        ),
                    )
                    for analyte in measurement.analytes
                ]
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.errors
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
            location_identifier=measurement.location_identifier,
            description=measurement.description,
            sample_role_type=measurement.sample_role_type,
            well_plate_identifier=measurement.well_plate_identifier,
        )

    def _get_calibration_aggregate_document(
        self, calibrations: list[Calibration] | None
    ) -> CalibrationAggregateDocument | None:
        if not calibrations:
            return None
        return CalibrationAggregateDocument(
            calibration_document=[
                CalibrationDocumentItem(
                    calibration_name=calibration.name,
                    calibration_report=calibration.report,
                    calibration_time=self.get_date_time(calibration.time),
                )
                for calibration in calibrations
            ]
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None

        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.feature)
                for error in errors
            ]
        )



from pathlib import Path
from typing import Any
import urllib.request

from allotropy.allotrope.schema_parser.path_util import (
    get_full_schema_path,
    get_schema_path_from_reference,
)


def _get_schema_from_reference(reference: str) -> str:
    return reference.split("#/$defs")[0]


def _get_references(schema: dict[str, Any]) -> set[str]:
    """Get all references ($ref:) from the given schema."""
    references = set()
    for key, value in schema.items():
        if key == "$ref":
            references.add(_get_schema_from_reference(value))
        elif isinstance(value, dict):
            references |= _get_references(value)
        elif isinstance(value, list):
            for v in value:
                references |= _get_references(v)
    return references


def download_schema(schema_url: str) -> Path:
    schema_path = get_full_schema_path(get_schema_path_from_reference(schema_url))
    if schema_path.exists():
        return schema_path
    if not schema_path.parent.exists():
        schema_path.parent.mkdir(parents=True, exist_ok=True)
    if not schema_url.startswith(("http:", "https:")):
        msg = f"Invald URL {schema_url}"
        raise ValueError(msg)
    # NOTE: S310 checks that you do not access a URL without checking it is a valid http url.
    # the code above does this, exactly as the documentation recommends, but it is still being
    # flagged, so ignore it.
    urllib.request.urlretrieve(schema_url, schema_path)  # noqa: S310
    return schema_path


def _download_references(references: set[str]) -> set[Path]:
    """Downloads all references to the schemas directory, returning the corresponding paths."""
    schema_paths: set[Path] = set()
    for reference in references:
        if reference.startswith("http"):
            schema_paths.add(download_schema(reference))
        else:
            if not Path(reference).exists():
                msg = f"Custom schema at path: '{reference}' does not exist, did you forget to add it?"
                raise AssertionError(msg)
            schema_paths.add(Path(reference))
    return schema_paths


def download_all_references_schemas_for_schema(schema: dict[str, Any]) -> set[Path]:
    references = _get_references(schema)
    return _download_references(references)



import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    assert_not_empty_df,
    df_to_series_data,
    parse_header_row,
    read_csv,
    read_excel,
    SeriesData,
)


class UnchainedLabsLunaticReader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"
    data: pd.DataFrame
    header: SeriesData

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        if named_file_contents.extension == "csv":
            data = read_csv(named_file_contents.contents, header=None)
        else:
            data = read_excel(named_file_contents.contents, header=None)

        assert_not_empty_df(data, "Unable to parse data from empty dataset.")

        # Find the header row by searching for a row with "Sample name".
        table_header_index = data.index.get_loc(
            data[(data == "Sample name").any(axis=1)].first_valid_index()
        )
        if table_header_index is None:
            msg = "Unable to find a table header row with 'Sample name'."
            raise AllotropeConversionError(msg)

        header_data = data[:table_header_index].dropna(how="all").T.dropna(how="all")
        data = parse_header_row(data[table_header_index:])
        # If the table header is the first row, or the non-empty data above the table header is a single title,
        # there is no metadata header, so attempt to read metadata from the first row of the table.
        # Otherwise, read in the metadata header and save it as a single series.
        if table_header_index == 0 or header_data.shape == (1, 1):
            self.header = df_to_series_data(data, index=0)
        else:
            self.header = df_to_series_data(
                parse_header_row(header_data).dropna(axis="columns")
            )

        # Fix column names in excel sheets with newlines/other whitespace.
        data.columns = (
            data.columns.astype(str)
            .str.replace("\n", " ")
            .str.replace("\r", "")
            .str.strip()
        )
        # Rows with no Sample name are assumed to be skipped measurements, and are dropped from the results.
        data = data.dropna(subset=["Sample name"])
        self.data = data.replace(np.nan, None)



from abc import abstractmethod


class Transformation:
    @abstractmethod
    def transform(self, data: tuple[float, ...]) -> tuple[float, ...]:
        pass


class Min2Sec(Transformation):
    def transform(self, data: tuple[float, ...]) -> tuple[float, ...]:
        return tuple(element * 60 for element in data)


class MScm2Sm(Transformation):
    def transform(self, data: tuple[float, ...]) -> tuple[float, ...]:
        return tuple(element / 10 for element in data)



from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    Measurement,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCubeComponent
from allotropy.parsers.cytiva_unicorn.constants import DEVICE_TYPE
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.generic import (
    UnicornMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.static_docs import (
    StaticDocs,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)


class ConcentrationMeasurement(UnicornMeasurement):
    @classmethod
    def create(
        cls,
        handler: UnicornZipHandler,
        elements: list[StrictXmlElement],
        static_docs: StaticDocs,
    ) -> Measurement:
        return cls.get_measurement(
            static_docs=static_docs,
            device_control_docs=[
                DeviceControlDoc(
                    device_type=DEVICE_TYPE,
                    solvent_conc_data_cube=cls.get_data_cube(
                        handler,
                        cls.filter_curve(elements, r"^Conc B$"),
                        DataCubeComponent(
                            type_=FieldComponentDatatype.float,
                            concept="solvent concentration",
                            unit="%",
                        ),
                    ),
                ),
            ],
        )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.example_weyland_yutani import constants
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass(frozen=True)
class BasicAssayInfo:
    protocol_id: str
    assay_id: str
    checksum: str | None

    @staticmethod
    def create(bottom: pd.DataFrame | None) -> BasicAssayInfo:
        checksum = (
            None
            if (bottom is None) or (bottom.iloc[0, 0] != "Checksum")
            else str(bottom.iloc[0, 1])
        )
        return BasicAssayInfo(
            # NOTE: in real code these values would be read from data
            protocol_id=constants.PROTOCOL_ID,
            assay_id=constants.ASSAY_ID,
            checksum=checksum,
        )


@dataclass(frozen=True)
class Instrument:
    serial_number: str
    nickname: str

    # TODO(tutorial): extract and fill in real values for serial number and nickname
    @staticmethod
    def create() -> Instrument:
        return Instrument(serial_number="", nickname="")


@dataclass(frozen=True)
class Result:
    col: str
    row: str
    value: float


@dataclass(frozen=True)
class Plate:
    number: str
    results: list[Result]

    @property
    def number_of_wells(self) -> int:
        return len(self.results)

    @staticmethod
    def create(df: pd.DataFrame | None) -> list[Plate]:
        if df is None:
            return []
        pivoted = df.T
        if pivoted.iloc[1, 0] != "A":
            msg = "Column header(s) not found."
            raise AllotropeConversionError(msg)
        stripped = pivoted.drop(0, axis="index").drop(0, axis="columns")
        rows, cols = stripped.shape
        stripped.index = [df.iloc[0, i + 1] for i in range(rows)]  # type: ignore
        stripped.columns = [str(int(df.iloc[i, 0])) for i in range(1, cols + 1)]  # type: ignore
        return [
            Plate(
                number="0",
                results=[
                    Result(col, row, float(stripped.loc[col, row]))
                    for col, row in stripped.stack().index
                ],
            )
        ]


def create_metadata(instrument: Instrument, file_path: str) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        model_number=instrument.serial_number,
        device_identifier=instrument.nickname,
    )


def create_measurement_group(
    plate: Plate, basic_assay_info: BasicAssayInfo
) -> MeasurementGroup:
    return MeasurementGroup(
        plate_well_count=plate.number_of_wells,
        analytical_method_identifier=basic_assay_info.protocol_id,
        experimental_data_identifier=basic_assay_info.assay_id,
        # TODO(tutorial): extract and return actual measurement time
        measurement_time="2022-12-31",
        measurements=[
            Measurement(
                type_=MeasurementType.FLUORESCENCE,
                device_type=constants.DEVICE_TYPE,
                identifier=random_uuid_str(),
                detection_type=constants.DETECTION_TYPE,
                sample_identifier=f"Plate {plate.number}",
                location_identifier=f"{result.col}{result.row}",
                fluorescence=result.value,
            )
            for result in plate.results
        ],
    )



from dataclasses import dataclass

import pandas as pd

from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Result,
    Well,
    WellItem,
    WellList,
)
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.values import assert_not_none


class GenotypingResult(Result):
    @classmethod
    def get_genotyping_determination_result(cls, target_data: SeriesData) -> str | None:
        return target_data.get(str, "Call")

    @classmethod
    def get_genotyping_determination_method_setting(
        cls, target_data: SeriesData
    ) -> float | None:
        return target_data.get(float, "Threshold")


class GenotypingWellItem(WellItem):
    @classmethod
    def get_result_class(cls) -> type[Result]:
        return GenotypingResult


class GenotypingWell(Well):
    @classmethod
    def get_well_item_class(cls) -> type[WellItem]:
        return GenotypingWellItem


@dataclass(frozen=True)
class GenotypingWellList(WellList):
    @classmethod
    def get_well_class(cls) -> type[Well]:
        return GenotypingWell

    @classmethod
    def get_well_result_data(cls, reader: DesignQuantstudioReader) -> pd.DataFrame:
        data = reader.get_non_empty_sheet(cls.get_data_sheet())
        genotyping_result = reader.get_non_empty_sheet("Genotyping Result")

        # The genotyping result data does not contain a target column
        # it can be constructed concatenating SNP assay column and the strings Allele 1/2
        rows = []
        for idx, row in genotyping_result.iterrows():
            snp_assay = assert_not_none(
                row.get("SNP Assay"),
                msg=f"Unable to get SNP Assay from Genotyping Result row '{idx}'.",
            )
            for allele in ["Allele 1", "Allele 2"]:
                new_row = row.copy()
                new_row["Target"] = f"{snp_assay}-{allele}"
                rows.append(new_row)

        return cls._add_data(
            data,
            extra_data=pd.DataFrame(rows).reset_index(drop=True),
            columns=[
                "Call",
            ],
        )



from pathlib import Path

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import (
    ExperimentType,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.qpcr import (
    CalculatedData,
    CalculatedDataItem,
    DataSource,
    Measurement,
    MeasurementGroup,
    Metadata,
    ProcessedData,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_data_creator import (
    _create_multicomponent_data_cubes,
    _create_processed_data_cubes,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    AmplificationData,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis import constants
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.creator import (
    Creator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Data,
    Header,
    MeltCurveData,
    Result,
    Well,
    WellItem,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.genotyping.creator import (
    GenotypingCreator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.melt_curve.creator import (
    MeltCurveCreator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.presence_absence.creator import (
    PresenceAbsenceCreator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.primary_analysis.creator import (
    PrimaryAnalysisCreator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.relative_standard_curve.creator import (
    RelativeStandardCurveCreator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.simple_primary_analysis.creator import (
    SimplePrimaryAnalysisCreator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.standard_curve.creator import (
    StandardCurveCreator,
)


def create_metadata(
    header: Header, file_path: str, experiment_type: ExperimentType
) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        device_identifier=header.device_identifier,
        model_number=header.model_number,
        device_serial_number=header.device_serial_number,
        data_system_instance_identifier=constants.DATA_SYSTEM_INSTANCE_IDENTIFIER,
        software_name=header.software_name,
        software_version=header.software_version,
        container_type=constants.CONTAINER_TYPE,
        device_type=constants.DEVICE_TYPE,
        experiment_type=experiment_type,
        measurement_method_identifier=header.measurement_method_identifier,
    )


def _create_processed_data(
    result: Result, amplification_data: AmplificationData | None
) -> ProcessedData:
    (
        normalized_reporter_data_cube,
        baseline_corrected_reporter_data_cube,
    ) = _create_processed_data_cubes(amplification_data)
    return ProcessedData(
        automatic_cycle_threshold_enabled_setting=result.automatic_cycle_threshold_enabled_setting,
        cycle_threshold_value_setting=result.cycle_threshold_value_setting,
        automatic_baseline_determination_enabled_setting=result.automatic_baseline_determination_enabled_setting,
        baseline_determination_start_cycle_setting=result.baseline_determination_start_cycle_setting,
        baseline_determination_end_cycle_setting=result.baseline_determination_end_cycle_setting,
        genotyping_determination_method_setting=result.genotyping_determination_method_setting,
        genotyping_determination_result=result.genotyping_determination_result,
        cycle_threshold_result=result.cycle_threshold_result,
        normalized_reporter_result=result.normalized_reporter_result,
        baseline_corrected_reporter_result=result.baseline_corrected_reporter_result,
        normalized_reporter_data_cube=normalized_reporter_data_cube,
        baseline_corrected_reporter_data_cube=baseline_corrected_reporter_data_cube,
    )


def _create_melt_curve_data_cube(
    melt_curve_raw_data: MeltCurveData | None,
) -> DataCube | None:
    if not melt_curve_raw_data:
        return None
    return DataCube(
        label="melting curve",
        structure_dimensions=[
            DataCubeComponent(FieldComponentDatatype.double, "temperature", "degC")
        ],
        structure_measures=[
            DataCubeComponent(FieldComponentDatatype.double, "fluorescence", "RFU"),
            DataCubeComponent(FieldComponentDatatype.double, "derivative", UNITLESS),
        ],
        dimensions=[melt_curve_raw_data.temperature],
        measures=[
            melt_curve_raw_data.fluorescence,
            melt_curve_raw_data.derivative,
        ],
    )


def _create_measurement(well: Well, well_item: WellItem, header: Header) -> Measurement:
    (
        reporter_dye_data_cube,
        passive_reference_dye_data_cube,
    ) = _create_multicomponent_data_cubes(
        well.multicomponent_data,
        well_item.reporter_dye_setting,
        header.passive_reference_dye_setting,
    )
    return Measurement(
        identifier=well_item.uuid,
        timestamp=header.measurement_time,
        target_identifier=well_item.target_dna_description,
        sample_identifier=well_item.sample_identifier,
        sample_role_type=well_item.sample_role_type,
        well_location_identifier=well_item.well_location_identifier,
        well_plate_identifier=header.barcode,
        total_cycle_number_setting=well_item.amplification_data.total_cycle_number_setting
        if well_item.amplification_data
        else None,
        pcr_detection_chemistry=header.pcr_detection_chemistry,
        reporter_dye_setting=well_item.reporter_dye_setting,
        quencher_dye_setting=well_item.quencher_dye_setting,
        passive_reference_dye_setting=header.passive_reference_dye_setting,
        processed_data=_create_processed_data(
            well_item.result, well_item.amplification_data
        ),
        reporter_dye_data_cube=reporter_dye_data_cube,
        passive_reference_dye_data_cube=passive_reference_dye_data_cube,
        melting_curve_data_cube=_create_melt_curve_data_cube(well_item.melt_curve_data),
    )


def create_measurement_groups(
    wells: list[Well], header: Header
) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(
            analyst=header.analyst,
            experimental_data_identifier=header.experimental_data_identifier,
            plate_well_count=header.plate_well_count,
            measurements=[
                _create_measurement(well, well_item, header)
                for well_item in well.items.values()
            ],
        )
        for well in wells
    ]


def create_calculated_data(data: Data) -> CalculatedData:
    return CalculatedData(
        reference_dna_description=data.reference_target,
        reference_sample_description=data.reference_sample,
        items=[
            CalculatedDataItem(
                identifier=calc_doc.uuid,
                name=calc_doc.name,
                value=calc_doc.value,
                unit=UNITLESS,
                data_sources=[
                    DataSource(
                        identifier=data_source.reference.uuid,
                        feature=data_source.feature,
                    )
                    for data_source in calc_doc.data_sources
                ],
            )
            for calc_doc in data.calculated_documents
        ],
    )


def create_data(reader: DesignQuantstudioReader) -> Data:
    possible_creators: list[type[Creator]] = [
        StandardCurveCreator,
        RelativeStandardCurveCreator,
        GenotypingCreator,
        MeltCurveCreator,
        PresenceAbsenceCreator,
        SimplePrimaryAnalysisCreator,
        PrimaryAnalysisCreator,
    ]

    raw_plugin_name = reader.header.get(str, "Plugin Name and Version")

    for creator in possible_creators:
        if creator.check_experiment_type(reader, raw_plugin_name):
            return creator.create(reader)

    msg = "Unable to infer experiment type from sheets in the input"
    raise AllotropeConversionError(msg)



from __future__ import annotations

import builtins
from collections.abc import Callable, Mapping, Sequence
from dataclasses import asdict, field, fields, is_dataclass, make_dataclass, MISSING
from enum import Enum
from types import GenericAlias, UnionType
from typing import (
    Any,
    cast,
    get_args,
    get_origin,
    TypeVar,
    Union,
)

from cattrs import Converter
from cattrs.errors import ClassValidationError
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
import numpy as np

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import (
    ProcessedDataDocumentItem,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
    TDimensionArray,
    TFunction,
    TMeasureArray,
)
from allotropy.allotrope.models.shared.definitions.units import HasUnit
from allotropy.allotrope.schema_parser.path_util import get_model_class_from_schema

SPECIAL_KEYS = {
    "manifest": "$asm.manifest",
    "field_asm_manifest": "$asm.manifest",
    "cube_structure": "cube-structure",
    "field_componentDatatype": "@componentDatatype",
    "field_asm_fill_value": "$asm.fill-value",
    "field_type": "@type",
    "field_index": "@index",
    "scan_position_setting__plate_reader_": "scan position setting (plate reader)",
    "detector_distance_setting__plate_reader_": "detector distance setting (plate reader)",
    "cell_type__cell_counter_": "cell type (cell counter)",
    "dead_cell_density__cell_counter_": "dead cell density (cell counter)",
    "average_dead_cell_diameter__cell_counter_": "average dead cell diameter (cell counter)",
    "viability__cell_counter_": "viability (cell counter)",
    "total_cell_density__cell_counter_": "total cell density (cell counter)",
    "viable_cell_density__cell_counter_": "viable cell density (cell counter)",
    "average_live_cell_diameter__cell_counter_": "average live cell diameter (cell counter)",
    "average_total_cell_diameter__cell_counter_": "average total cell diameter (cell counter)",
    "total_cell_diameter_distribution__cell_counter_": "total cell diameter distribution (cell counter)",
    "viable_cell_count__cell_counter_": "viable cell count (cell counter)",
    "total_cell_count__cell_counter_": "total cell count (cell counter)",
    "autosampler_injection_volume_setting__chromatography_": "autosampler injection volume setting (chromatography)",
    "capacity_factor__chromatography_": "capacity factor (chromatography)",
    "peak_selectivity__chromatography_": "peak selectivity (chromatography)",
    "peak_width_at_4_4___of_height": "peak width at 4.4 % of height",
    "peak_width_at_13_4___of_height": "peak width at 13.4 % of height",
    "peak_width_at_32_4___of_height": "peak width at 32.4 % of height",
    "peak_width_at_60_7___of_height": "peak width at 60.7 % of height",
    "peak_width_at_5___of_height": "peak width at 5 % of height",
    "peak_width_at_10___of_height": "peak width at 10 % of height",
    "statistical_skew__chromatography_": "statistical skew (chromatography)",
    "asymmetry_factor_measured_at_5___height": "asymmetry factor measured at 5 % height",
    "asymmetry_factor_measured_at_10___height": "asymmetry factor measured at 10 % height",
    "asymmetry_factor_squared_measured_at_10___height": "asymmetry factor squared measured at 10 % height",
    "asymmetry_factor_squared_measured_at_4_4___height": "asymmetry factor squared measured at 4.4 % height",
    "asymmetry_factor_measured_at_4_4___height": "asymmetry factor measured at 4.4 % height",
    "number_of_theoretical_plates__chromatography_": "number of theoretical plates (chromatography)",
    "number_of_theoretical_plates_measured_at_60_7___of_peak_height": "number of theoretical plates measured at 60.7 % of peak height",
    "number_of_theoretical_plates_measured_at_32_4___of_peak_height": "number of theoretical plates measured at 32.4 % of peak height",
    "number_of_theoretical_plates_measured_at_13_4___of_peak_height": "number of theoretical plates measured at 13.4 % of peak height",
    "number_of_theoretical_plates_measured_at_4_4___of_peak_height": "number of theoretical plates measured at 4.4 % of peak height",
    "number_of_theoretical_plates_by_peak_width_at_half_height__JP14_": "number of theoretical plates by peak width at half height (JP14)",
    "confidence_interval__95__": "confidence interval (95%)",
    "co2_saturation": "CO2 saturation",
    "o2_saturation": "O2 saturation",
    "pco2": "pCO2",
    "po2": "pO2",
}
SPECIAL_KEYS_INVERSE: dict[str, str] = dict(
    cast(tuple[str, str], reversed(item)) for item in SPECIAL_KEYS.items()
)


DICT_KEY_TO_MODEL_KEY_REPLACEMENTS = {
    ".": "_POINT_",
    "-": "_DASH_",
    "°": "_DEG_",
    "/": "_SLASH_",
    "\\": "_BSLASH_",
    "(": "_OPAREN_",
    ")": "_CPAREN_",
    "%": "_PERCENT_",
    ":": "_COLON_",
    # NOTE: this MUST be at the end, or it will break other key replacements.
    " ": "_",
}


PRIMITIVE_TYPES = (
    bool,
    int,
    float,
    str,
    type(None),
    InvalidJsonFloat,
    np.float64,
    np.int64,
)

ModelClass = TypeVar("ModelClass")


def add_custom_information_document(
    model: ModelClass, custom_info_doc: Any
) -> ModelClass:
    if not custom_info_doc:
        return model

    # Convert to a dictionary first, so we can clean up values.
    if is_dataclass(custom_info_doc):
        custom_info_dict = asdict(custom_info_doc)
    elif isinstance(custom_info_doc, dict):
        custom_info_dict = custom_info_doc
    else:
        msg = f"Invalid custom_info_doc: {custom_info_doc}"
        raise ValueError(msg)

    # Remove None and {"value": None, "unit"...} values
    cleaned_dict = {}
    for key, value in custom_info_dict.items():
        if value is None:
            continue
        if isinstance(value, dict) and "value" in value and value["value"] is None:
            continue
        cleaned_dict[key] = value

    # If dict is empty after cleaning, do not attach.
    if not cleaned_dict:
        return model

    custom_info_doc = structure_custom_information_document(
        cleaned_dict, "custom information document"
    )

    model.custom_information_document = custom_info_doc  # type: ignore
    return model


def _convert_model_key_to_dict_key(key: str) -> str:
    key = SPECIAL_KEYS.get(key, key)
    for dict_val, model_val in DICT_KEY_TO_MODEL_KEY_REPLACEMENTS.items():
        key = key.replace(model_val, dict_val)
    return key


def _convert_dict_to_model_key(key: str) -> str:
    key = SPECIAL_KEYS_INVERSE.get(key, key)
    for dict_val, model_val in DICT_KEY_TO_MODEL_KEY_REPLACEMENTS.items():
        key = key.replace(dict_val, model_val)
    return key


def _validate_structuring(val: Any, model: Any) -> None:
    """Validate that all keys in val are stored in model."""
    if isinstance(val, list):
        if not isinstance(model, list):
            raise AssertionError()
        for list_value, model_list_value in zip(val, model, strict=True):
            _validate_structuring(list_value, model_list_value)
    if not isinstance(val, dict):
        return

    for key, value in val.items():
        model_key = _convert_dict_to_model_key(key)
        # If the key is unit, and this is a unit model, ensure the unit is correct.
        if key == "unit" and isinstance(model, HasUnit):
            unit_field = next(field for field in fields(model) if field.name == "unit")
            if not value == unit_field.default:
                raise AssertionError()
        # If the value itself is None, just assert that the key is in the model.
        if value is None:
            if not hasattr(model, model_key):
                raise AssertionError()
            continue

        model_val = getattr(model, model_key, None)
        if model_val is None:
            raise AssertionError()

        _validate_structuring(value, model_val)


def register_data_cube_hooks(converter: Converter) -> None:
    def structure_dimension_array(val: Any, _: Any) -> TDimensionArray | TFunction:
        if isinstance(val, list):
            return val
        return converter.structure(val, TFunction)

    converter.register_structure_hook(
        TDimensionArray | TFunction, structure_dimension_array
    )
    converter.register_structure_hook(TMeasureArray, lambda val, _: val)


# TODO: this code is copied from cattrs 24.1.0. We currently pin to 23.1.2 because some other libraries that
# allotropy is currently used with pin cattrs (i.e. ddtrace). When possible, upgrade to cattrs>=24.1.0 and
# import is_sequence from cattrs.cols
def is_subclass(obj: type, bases: builtins._ClassInfo) -> bool:
    """A safe version of issubclass (won't raise)."""
    try:
        return issubclass(obj, bases)
    except TypeError:
        return False


def is_sequence(type_: Any) -> bool:
    origin = getattr(type_, "__origin__", None)
    return type_ in (list, tuple) or (
        type_.__class__ in (GenericAlias, type)
        and origin
        and (origin not in (Union, tuple) and is_subclass(origin, Sequence))
        or (origin is tuple and type_.__args__[1] is ...)
    )


def register_dataclass_union_hooks(converter: Converter) -> None:
    # Handles any union of dataclass, lists of dataclasses, and primitive values.
    # First checks if the value is a list, and if so tries to parse with any of the list types.
    # Then checks if the value is a primitive value or None, and if so returns that.
    # Then tries structuring with each specified dataclass, if any.
    def _is_valid(arg: Any) -> bool:
        if is_sequence(arg):
            return all(is_dataclass(sub) for sub in arg.__args__)
        if is_dataclass(arg):
            return True
        if arg in PRIMITIVE_TYPES:
            return True
        return False

    def is_dataclass_union(val: Any) -> bool:
        if get_origin(val) not in (Union, UnionType):
            return False
        args = set(get_args(val))
        return all(_is_valid(arg) for arg in args)

    def dataclass_union_structure_fn(
        cls: Any,
    ) -> Callable[[dict[str, Any] | str | None, Any], Any | None]:
        def structure_item(val: dict[str, Any] | str | None, _: Any) -> Any | None:
            if isinstance(val, list):
                valid_models = []
                for subcls in get_args(cls):
                    if not is_sequence(subcls):
                        continue
                    # I don't think this should be possible, but check and raise a readable error just in case.
                    if len(subcls.__args__) > 1:
                        msg = (
                            f"Encountered a list type with more than one arg: {subcls}!"
                        )
                        raise AssertionError(msg)
                    try:
                        valid_models.append(
                            [converter.structure(v, subcls.__args__[0]) for v in val]
                        )
                    except (ClassValidationError, TypeError):
                        pass

                if len(valid_models) == 1:
                    return valid_models[0]
                elif len(valid_models) > 1:
                    for model in valid_models:
                        try:
                            _validate_structuring(val, model)
                            return model
                        except AssertionError:
                            pass

            if type(val) in PRIMITIVE_TYPES:
                return val
            valid_models = []
            for subcls in get_args(cls):
                if not is_dataclass(subcls):
                    continue
                try:
                    valid_models.append(converter.structure(val, subcls))
                except ClassValidationError:
                    pass

            if len(valid_models) == 1:
                return valid_models[0]
            elif len(valid_models) > 1:
                for model in valid_models:
                    try:
                        _validate_structuring(val, model)
                        return model
                    except AssertionError:
                        pass

            msg = f"Failed to structure value {val} with type {cls}"
            raise ValueError(msg)

        return structure_item

    converter.register_structure_hook_factory(
        is_dataclass_union, dataclass_union_structure_fn
    )


def structure_custom_information_document(val: dict[str, Any], name: str) -> Any:
    structured_dict = {}
    for key, value in val.items():
        structured_value = value
        if isinstance(value, list):
            structured_value = [
                structure_custom_information_document(v, key) for v in value
            ]
        elif isinstance(value, dict):
            structured_value = structure_custom_information_document(value, key)
        structured_dict[_convert_dict_to_model_key(key)] = structured_value

    name = name.title().replace(" ", "")
    return make_dataclass(
        name, ((k, type(v), field(default=None)) for k, v in structured_dict.items())
    )(**structured_dict)


def _create_should_omit_function(
    cls: Any, parent_cls: Any | None = None, field_name: str | None = None
) -> Callable[[str, Any], bool]:
    required_keys = {a.name for a in fields(cls) if a.default == MISSING}

    def should_omit(k: str, v: Any) -> bool:
        if k in required_keys:
            return False
        if field_name in EMPTY_VALUE_CLASS_AND_FIELD.get(parent_cls, set()):
            return v is None and k != "value"
        return v is None

    return should_omit


def _unstructure_value(value: Any) -> Any:
    return value.value if isinstance(value, Enum) else value


def unstructure_custom_information_document(model: Any) -> dict[str, Any]:
    should_omit = _create_should_omit_function(model)

    def dict_factory(kv_pairs: Sequence[tuple[str, Any]]) -> dict[str, Any]:
        return {
            _convert_model_key_to_dict_key(key): _unstructure_value(value)
            for key, value in kv_pairs
            if not should_omit(key, value)
        }

    return asdict(model, dict_factory=dict_factory)


def register_dataclass_hooks(converter: Converter) -> None:
    def dataclass_structure_fn(cls: Any) -> Callable[[Any, Any], Any | None]:
        structure_fn = make_dict_structure_fn(
            cls,
            converter,
            **{
                a.name: override(rename=_convert_model_key_to_dict_key(a.name))
                for a in fields(cls)
            },
        )

        def structure_item(val: Any, _: Any) -> Any | None:
            if val is None:
                return None
            structured = structure_fn(val, _)
            if isinstance(val, dict) and "custom information document" in val:
                structured.custom_information_document = (
                    structure_custom_information_document(
                        val["custom information document"],
                        "custom information document",
                    )
                )
            return structured

        return structure_item

    converter.register_structure_hook_factory(is_dataclass, dataclass_structure_fn)


QPCR_NULLABLE_VALUE_CLASSES: dict[Any, set[str]] = {
    ProcessedDataDocumentItem: {"cycle_threshold_result"}
}

EMPTY_VALUE_CLASS_AND_FIELD = {
    **QPCR_NULLABLE_VALUE_CLASSES,
}


def register_unstructure_hooks(converter: Converter) -> None:
    unstructure_fn_cache = {}

    def unstructure_dataclass_fn(
        cls: Any, parent_cls: Any | None = None, field_name: str | None = None
    ) -> Callable[[Any], dict[str, Any]]:
        should_omit = _create_should_omit_function(cls, parent_cls, field_name)

        def unstructure(obj: Any) -> Any:
            # Break out of dataclass recursion by calling back to converter.unstructure
            if not is_dataclass(obj):
                return converter.unstructure(obj)

            dataclass_dict = {
                _convert_model_key_to_dict_key(k): v
                for k, v in make_unstructure_fn(type(obj))(obj).items()
                if not should_omit(k, v)
            }
            if hasattr(obj, "custom_information_document"):
                dataclass_dict[
                    "custom information document"
                ] = unstructure_custom_information_document(
                    obj.custom_information_document
                )
            return dataclass_dict

        # This custom unstructure function overrides the unstruct_hook. We need to do this at this level
        # because we need to know both the parent class and the field name at the same time to create the
        # should_omit function.
        def make_unstructure_fn(subcls: Any) -> Callable[[Any], dict[str, Any]]:
            if (cls, subcls) not in unstructure_fn_cache:
                unstructure_fn_cache[(cls, subcls)] = make_dict_unstructure_fn(
                    subcls,
                    converter,
                    **{
                        a.name: override(
                            unstruct_hook=unstructure_dataclass_fn(subcls, cls, a.name)
                        )
                        for a in fields(cls)
                    },
                )
            return unstructure_fn_cache[(cls, subcls)]

        return unstructure

    converter.register_unstructure_hook_factory(is_dataclass, unstructure_dataclass_fn)


def setup_converter() -> Converter:
    converter = Converter()
    register_data_cube_hooks(converter)
    register_dataclass_union_hooks(converter)
    register_dataclass_hooks(converter)
    register_unstructure_hooks(converter)
    return converter


CONVERTER = setup_converter()


def unstructure(model: Any) -> dict[str, Any]:
    return cast(dict[str, Any], CONVERTER.unstructure(model))


def structure(asm: Mapping[str, Any], model_class: Any | None = None) -> Any:
    model_class = model_class or get_model_class_from_schema(asm)
    return CONVERTER.structure(asm, model_class)



from __future__ import annotations

from io import BytesIO

from allotropy.parsers.cytiva_unicorn.reader.zip_handler import (
    ZipHandler,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)


class UnicornZipHandler(ZipHandler):
    def get_zip(self, inner_path: str) -> UnicornZipHandler:
        return UnicornZipHandler(self.get_file(inner_path))

    def get_zip_from_pattern(self, pattern: str) -> UnicornZipHandler:
        return self.get_zip(self.get_inner_path(pattern))

    @classmethod
    def create_from_path(cls, path: str) -> UnicornZipHandler:
        with open(path, "rb") as f:
            data = f.read()
        return UnicornZipHandler(data=BytesIO(data))

    def filter_xml_metadata(self, stream: BytesIO) -> BytesIO:
        data = stream.read()
        lower_symbol, upper_symbol = 32, 126

        start = 0
        for idx, element in enumerate(data):
            if lower_symbol <= int(element) <= upper_symbol:
                start = idx
                break

        end = len(data)
        for idx, element in enumerate(reversed(data)):
            if lower_symbol <= int(element) <= upper_symbol:
                end -= idx
                break

        return BytesIO(data[start:end])

    def get_system_data(self) -> StrictXmlElement:
        system_data = self.get_zip_from_pattern("SystemData.zip$")
        b_stream = system_data.get_file_from_pattern("^Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_results(self) -> StrictXmlElement:
        b_stream = self.get_file_from_pattern("Result.xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_instrument_config_data(self) -> StrictXmlElement:
        instrument_regex = "InstrumentConfigurationData.zip$"
        instrument_config_data = self.get_zip_from_pattern(instrument_regex)
        b_stream = instrument_config_data.get_file_from_pattern("^Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_evaluation_log(self) -> StrictXmlElement:
        b_stream = self.get_file_from_pattern("EvaluationLog.xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_chrom_1(self) -> StrictXmlElement:
        b_stream = self.get_file_from_pattern("Chrom.1.Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)

    def get_column_type_data(self) -> StrictXmlElement:
        column_type_data = self.get_zip_from_pattern("ColumnTypeData.zip$")
        b_stream = column_type_data.get_file_from_pattern("^Xml$")
        raw_content = self.filter_xml_metadata(b_stream).read()
        return StrictXmlElement.create_from_bytes(raw_content)



from __future__ import annotations

from allotropy.allotrope.models.adm.cell_counting.rec._2024._09.cell_counting import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_reader import ViCellBluReader
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_structure import (
    create_measurement_group,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.pandas import map_rows
from allotropy.parsers.vendor_parser import VendorParser


class ViCellBluParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Beckman Vi-Cell BLU"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = ViCellBluReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        return Data(
            create_metadata(named_file_contents.original_file_path),
            map_rows(
                ViCellBluReader.read(named_file_contents), create_measurement_group
            ),
        )



from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.converters import (
    Converter,
    FloatConverter,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.transformations import (
    Transformation,
)


class DataCubeReader:
    def __init__(
        self,
        handler: UnicornZipHandler,
        name: str,
        transformation: Transformation | None,
    ):
        self.transformation = transformation

        self.data_stream = handler.get_file_from_pattern(f"CoordinateData.{name}$")
        self.type_stream = handler.get_file_from_pattern(
            f"CoordinateData.{name}DataType$"
        )

        self.data = self.data_stream.read()
        self.type = self.type_stream.read().decode()

    def get_converter(self) -> Converter:
        if self.type.startswith("System.Single"):
            return FloatConverter()

        msg = f"Unable to parse data cube with binary data in format {self.type}"
        raise AllotropeConversionError(msg)

    def get_data(self) -> tuple[float, ...]:
        data = self.get_converter().convert(self.data)
        return self.transformation.transform(data) if self.transformation else data



# generated by datamodel-codegen:
#   filename:  electrophoresis.schema.json
#   timestamp: 2024-08-13T06:10:50+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueKiloDalton,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValuePercent,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TClass,
    TDatacube,
    TDateTimeStampValue,
    TIntegerValue,
    TStatisticDatumRole,
    TStringValue,
    TUnit,
)


@dataclass(kw_only=True)
class DiagnosticTraceDocumentItem:
    description: Any


@dataclass(kw_only=True)
class DiagnosticTraceAggregateDocument:
    diagnostic_trace_document: list[DiagnosticTraceDocumentItem] | None = None


@dataclass(kw_only=True)
class AdmCoreREC202403ManifestSchema:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: str | None = None
    field_type: str | None = None
    shapes: list[str] | None = None


@dataclass(kw_only=True)
class ErrorDocumentItem:
    error: TStringValue
    error_feature: TStringValue | None = None


@dataclass(kw_only=True)
class ErrorAggregateDocument:
    error_document: list[ErrorDocumentItem] | None = None


@dataclass(kw_only=True)
class ImageDocumentItem:
    index: TIntegerValue | None = None
    experimental_data_identifier: TStringValue | None = None


@dataclass(kw_only=True)
class ImageAggregateDocument:
    image_document: list[ImageDocumentItem] | None = None


@dataclass(kw_only=True)
class StatisticsDocumentItem:
    statistical_feature: TClass


@dataclass(kw_only=True)
class StatisticsAggregateDocument:
    statistics_document: list[StatisticsDocumentItem] | None = None


@dataclass(kw_only=True)
class TQuantityValueModel:
    value: float
    unit: TUnit
    has_statistic_datum_role: TStatisticDatumRole | None = None
    field_type: TClass | None = None


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue | None = None
    description: Any | None = None
    brand_name: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    device_document: list[DeviceDocumentItem] | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    data_system_instance_identifier: TStringValue | None = None
    file_name: TStringValue | None = None
    UNC_path: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    detection_type: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    model_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    field_index: int | None = None
    detector_wavelength_setting: TQuantityValueNanometer | None = None
    detector_bandwidth_setting: TQuantityValueNanometer | None = None
    excitation_wavelength_setting: TQuantityValueNanometer | None = None
    excitation_bandwidth_setting: TQuantityValueNanometer | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue
    description: Any | None = None
    batch_identifier: TStringValue | None = None
    sample_role_type: TClass | None = None
    written_name: TStringValue | None = None
    location_identifier: TStringValue | None = None


@dataclass(kw_only=True)
class PeakItem:
    peak_identifier: TStringValue
    peak_name: TStringValue | None = None
    written_name: TStringValue | None = None
    peak_height: TQuantityValueMilliAbsorbanceUnit | TQuantityValueRelativeFluorescenceUnit | None = (
        None
    )
    relative_peak_height: TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    peak_start: TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    peak_end: TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    peak_position: TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    peak_area: TQuantityValueUnitless | None = None
    relative_peak_area: TQuantityValuePercent | None = None
    relative_corrected_peak_area: TQuantityValuePercent | None = None
    comment: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class PeakList:
    peak: list[PeakItem]


@dataclass(kw_only=True)
class DataRegionDocumentItem:
    region_identifier: TStringValue
    region_name: TStringValue | None = None
    written_name: TStringValue | None = None
    region_start: TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    region_end: TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    region_area: TQuantityValueUnitless | None = None
    relative_region_area: TQuantityValuePercent | None = None
    comment: TStringValue | None = None


@dataclass(kw_only=True)
class DataRegionAggregateDocument:
    data_region_document: list[DataRegionDocumentItem]


@dataclass(kw_only=True)
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue
    field_index: int | None = None


@dataclass(kw_only=True)
class DataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem]


@dataclass(kw_only=True)
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValueModel
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    calculated_data_identifier: TStringValue | None = None
    calculation_description: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class CalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass(kw_only=True)
class ProcessedDataDocumentItem:
    peak_list: PeakList | None = None
    data_processing_document: dict[str, Any] | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    processed_data_identifier: TStringValue | None = None
    field_index: int | None = None
    derived_electropherogram_data_cube: TDatacube | None = None
    data_region_aggregate_document: DataRegionAggregateDocument | None = None


@dataclass(kw_only=True)
class MeasurementDocumentItem:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    compartment_temperature: TQuantityValueDegreeCelsius | None = None
    electropherogram_data_cube: TDatacube | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocumentItem]
    diagnostic_trace_aggregate_document: DiagnosticTraceAggregateDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None


@dataclass(kw_only=True)
class ElectrophoresisDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument
    analyst: TStringValue | None = None
    submitter: TStringValue | None = None
    analytical_method_identifier: TStringValue | None = None
    method_version: TStringValue | None = None
    experimental_data_identifier: TStringValue | None = None
    experiment_type: TStringValue | None = None


@dataclass(kw_only=True)
class ElectrophoresisAggregateDocument:
    data_system_document: DataSystemDocument
    electrophoresis_document: list[ElectrophoresisDocumentItem]
    device_system_document: DeviceSystemDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None


@dataclass(kw_only=True)
class Model:
    electrophoresis_aggregate_document: ElectrophoresisAggregateDocument
    field_asm_manifest: AdmCoreREC202403ManifestSchema | str



from __future__ import annotations

import numpy as np
import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    assert_not_empty_df,
    df_to_series_data,
    parse_header_row,
    read_multisheet_excel,
    SeriesData,
    split_header_and_data,
)
from allotropy.parsers.utils.values import (
    assert_not_none,
)


class DesignQuantstudioReader:
    SUPPORTED_EXTENSIONS = "xlsx,xls"
    header: SeriesData
    data: dict[str, pd.DataFrame]

    @staticmethod
    def create(named_file_contents: NamedFileContents) -> DesignQuantstudioReader:
        return DesignQuantstudioReader(
            read_multisheet_excel(
                named_file_contents.contents,
                header=None,
                engine="calamine",
            )
        )

    def __init__(self, raw_contents: dict[str, pd.DataFrame]) -> None:
        contents = {
            str(name): df.replace(np.nan, None) for name, df in raw_contents.items()
        }
        self.header = self._get_header(contents)
        self.data = self._get_data(contents)

    def _get_header(self, contents: dict[str, pd.DataFrame]) -> SeriesData:
        sheet = assert_not_none(
            contents.get("Results"),
            msg="Unable to find 'Results' sheet.",
        )
        df, _ = split_header_and_data(sheet, lambda row: row[0] is None)
        return df_to_series_data(parse_header_row(df.T))

    def _get_data(self, contents: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        data_structure = {}
        for name, sheet in contents.items():
            _, data = split_header_and_data(sheet, lambda row: row[0] is None)
            data_structure[name] = parse_header_row(data.replace(np.nan, None))
        return data_structure

    def has_sheet(self, sheet_name: str) -> bool:
        return sheet_name in self.data

    def get_sheet_or_none(self, sheet_name: str) -> pd.DataFrame | None:
        return self.data.get(sheet_name)

    def get_non_empty_sheet_or_none(self, sheet_name: str) -> pd.DataFrame | None:
        sheet = self.get_sheet_or_none(sheet_name)
        return None if sheet is None or sheet.empty else sheet

    def get_sheet(self, sheet_name: str) -> pd.DataFrame:
        return assert_not_none(
            self.get_sheet_or_none(sheet_name),
            msg=f"Unable to find '{sheet_name}' sheet in file.",
        )

    def get_non_empty_sheet(self, sheet_name: str) -> pd.DataFrame:
        return assert_not_empty_df(
            self.get_sheet(sheet_name),
            msg=f"sheet '{sheet_name}' is empty.",
        )



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.ctl_immunospot.ctl_immunospot_structure import (
    AssayData,
    create_measurement_groups,
    create_metadata,
    Header,
)
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CtlImmunospotParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "CTL ImmunoSpot"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "txt"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = LinesReader.create(named_file_contents)
        header = Header.create(reader)
        reader.drop_empty()
        reader.drop_until_empty()  # ignore assay info
        reader.drop_empty()
        assay_data = AssayData.create(reader)

        return Data(
            create_metadata(header),
            create_measurement_groups(assay_data, header),
        )



from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ContainerType

DEVICE_TYPE = "qPCR"
PRODUCT_MANUFACTURER = "ThermoFisher Scientific"
DATA_SYSTEM_INSTANCE_IDENTIFIER = "localhost"
CONTAINER_TYPE = ContainerType.qPCR_reaction_block



DISPLAY_NAME = "Tecan Magellan"
DEVICE_TYPE = "plate reader"
SOFTWARE_NAME = "Magellan"
PRODUCT_MANUFACTURER = "Tecan"

MEASUREMENT_TIME_REGEX = (
    r"Date of measurement: ([0-9\-]+)/Time of measurement: ([0-9\:]+)"
)



from typing import ClassVar

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ExperimentType
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.creator import (
    Creator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Data,
    Header,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.genotyping.structure import (
    GenotypingWellList,
)


class GenotypingCreator(Creator):
    PLUGIN_REGEX: ClassVar[str] = r"Genotyping"
    EXPECTED_SHEETS: ClassVar[list[str]] = [
        "Genotyping Result",
    ]

    @classmethod
    def create(cls, reader: DesignQuantstudioReader) -> Data:
        header = Header.create(reader.header)
        wells = GenotypingWellList.create(reader, header)
        return Data(
            header,
            wells,
            experiment_type=ExperimentType.genotyping_qPCR_experiment,
            calculated_documents=[],
            reference_target=None,
            reference_sample=None,
        )



from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType

DETECTION_TYPE = "luminescence"
DEVICE_TYPE = "luminescence detector"
SOFTWARE_NAME = "MSD Workbench"

SAMPLE_ROLE_TYPE_MAPPING = {
    "c": SampleRoleType.control_sample_role,
    "s": SampleRoleType.standard_sample_role,
    "b": SampleRoleType.blank_role,
    "u": SampleRoleType.unknown_sample_role,
}

POSSIBLE_WELL_COUNTS = [1, 2, 4, 6, 8, 12, 24, 48, 72, 96, 384, 1536, 3456]



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import parse_header_row, read_excel, SeriesData


class BeckmanPharmspecReader:
    SUPPORTED_EXTENSIONS = "xlsx"
    data: pd.DataFrame
    header: SeriesData

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        df = read_excel(named_file_contents.contents, header=None, engine="calamine")

        """
        Find the data in the raw dataframe. We identify the boundary of the data
        by finding the index first row which contains the word 'Particle' and ending right before
        the index of the first row containing 'Approver'.
        """
        start = df[df[1].str.contains("Particle", na=False)].index.values[0]
        end = df[df[0].str.contains("Approver_", na=False)].index.values[0] - 1

        # The header data is everything up to the start of the data.
        # It is stored in two columns spread over the first 6 columns.
        raw_header = df.loc[: start - 1].T
        header_data = pd.concat([raw_header.loc[2], raw_header.loc[5]])
        header_columns = pd.concat([raw_header.loc[0], raw_header.loc[3]])
        header_data.index = pd.Index(header_columns)
        self.header = SeriesData(header_data)

        data = df.loc[start:end]
        data = parse_header_row(data)
        data["Run No."] = data["Run No."].ffill()
        self.data = data.dropna(subset="Particle Size(µm)")



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol, TypeVar

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
    TDimensionArray,
    TFunction,
    TMeasureArray,
)
from allotropy.exceptions import AllotropeConversionError


@dataclass(frozen=True)
class DataCubeComponent:
    type_: FieldComponentDatatype
    concept: str
    unit: str


@dataclass(frozen=True)
class DataCube:
    label: str
    structure_dimensions: list[DataCubeComponent]
    structure_measures: list[DataCubeComponent]
    dimensions: Sequence[Sequence[float] | Sequence[str] | Sequence[bool]]
    measures: Sequence[
        Sequence[float | None] | Sequence[str | None] | Sequence[bool | None]
    ]


class DataCubeProtocol(Protocol):
    def __init__(
        self,
        label: str,
        cube_structure: TDatacubeStructure,
        data: TDatacubeData,
    ):
        pass


DataCubeType = TypeVar("DataCubeType", bound=DataCubeProtocol)


T = TypeVar("T", bound=float | str | bool)


def _is_type(type_: type[T], value: float | str | int) -> bool:
    return isinstance(value, type_) or type_ is float and isinstance(value, int)


def _get_typed_dimension(
    type_: type[T], values: Sequence[float | str | bool]
) -> list[T] | None:
    # Allow ints or floats for float list.
    result: list[T] = [
        # Typing is not smart enough to tell that calling type_(value) results in a value of type: type_,
        # which I just can't deal with.
        type_(value)  # type: ignore[misc]
        for value in values
        if _is_type(type_, value)
    ]
    return result if len(result) == len(values) else None


def _get_dimensions(
    dimensions: Sequence[Sequence[float] | Sequence[str] | Sequence[bool]],
) -> list[TDimensionArray | TFunction]:
    result: list[TDimensionArray | TFunction] = []
    for dimension in dimensions:
        float_list = _get_typed_dimension(float, dimension)
        str_list = _get_typed_dimension(str, dimension)
        bool_list = _get_typed_dimension(bool, dimension)
        # NOTE: given the input types, this should be impossible, however, typing does not believe this, so we
        # add this check. This allows us to safely put "or []" at the end of result.append, confident that all
        # 3 lists are only falsy if the input is empty.
        if float_list is None and str_list is None and bool_list is None:
            msg = f"Unable to extract a TDimensionArray from datacube dimension: {dimension}"
            raise AllotropeConversionError(msg)
        result.append(float_list or str_list or bool_list or [])
    return result


def _get_typed_measure(
    type_: type[T], values: Sequence[float | str | bool | None]
) -> list[T | None] | None:
    result: list[T | None] = [
        # Typing is not smart enough to tell that calling type_(value) results in a value of type: type_,
        # which I just can't deal with.
        None if value is None else type_(value)  # type: ignore[misc]
        for value in values
        if value is None or _is_type(type_, value)
    ]
    return result if len(result) == len(values) else None


def _get_measures(
    measures: Sequence[
        Sequence[float | None] | Sequence[str | None] | Sequence[bool | None]
    ],
) -> list[TMeasureArray]:
    result: list[TMeasureArray] = []
    for measure in measures:
        float_list = _get_typed_measure(float, measure)
        str_list = _get_typed_measure(str, measure)
        bool_list = _get_typed_measure(bool, measure)
        # NOTE: given the input types, this should be impossible, however, typing does not believe us, so we
        # add
        # this check. This allows us to safely put "or []" at the end of result.append, confident that all
        # 3 lists are only falsy if the input is empty.
        if float_list is None and str_list is None and bool_list is None:
            msg = f"Unable to extract a TMeasureArray from datacube measure: {measure}"
            raise AllotropeConversionError(msg)
        result.append(float_list or str_list or bool_list or [])
    return result


def get_data_cube(
    data_cube: DataCube | None, data_cube_class: type[DataCubeType]
) -> DataCubeType | None:
    if data_cube is None:
        return None
    return data_cube_class(
        label=data_cube.label,
        cube_structure=TDatacubeStructure(
            dimensions=[
                TDatacubeComponent(
                    field_componentDatatype=structure_dim.type_,
                    concept=structure_dim.concept,
                    unit=structure_dim.unit,
                )
                for structure_dim in data_cube.structure_dimensions
            ],
            measures=[
                TDatacubeComponent(
                    field_componentDatatype=structure_dim.type_,
                    concept=structure_dim.concept,
                    unit=structure_dim.unit,
                )
                for structure_dim in data_cube.structure_measures
            ],
        ),
        data=TDatacubeData(
            dimensions=_get_dimensions(data_cube.dimensions),
            measures=_get_measures(data_cube.measures),
        ),
    )



""" Constants file for Thermo SkanIt"""
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType

SAMPLE_ROLE_MAPPINGS = {
    "Un": SampleRoleType.unknown_sample_role,
    "Std": SampleRoleType.standard_sample_role,
    "Blank": SampleRoleType.blank_role,
    "Ctrl": SampleRoleType.control_sample_role,
}
DEVICE_TYPE = "plate reader"



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
import re
from typing import Any

import pandas as pd

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import ContainerType
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import (
    CalculatedDataItem,
    DataSource,
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.appbio_absolute_q.constants import (
    AGGREGATION_LOOKUP,
    AggregationType,
    BRAND_NAME,
    CALCULATED_DATA_REFERENCE,
    CalculatedDataSource,
    CONCENTRATION_COLUMNS,
    DEVICE_TYPE,
    get_dye_settings,
    PLATE_WELL_COUNT,
    POSSIBLE_DYE_SETTING_LENGTHS,
    PRODUCT_MANUFACTURER,
    SOFTWARE_NAME,
)
from allotropy.parsers.constants import (
    DEFAULT_EPOCH_TIMESTAMP,
    NEGATIVE_ZERO,
    NOT_APPLICABLE,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass
class CalculatedItem:
    identifier: str
    name: str
    value: float
    unit: str
    source: CalculatedDataSource
    source_features: list[str]
    column: str

    def get_data_sources(
        self, measurement_ids: list[str], calculated_data_ids: dict[str, str]
    ) -> list[DataSource]:
        if self.source == CalculatedDataSource.CALCULATED_DATA:
            return [
                DataSource(calculated_data_ids[source_feature], source_feature)
                for source_feature in self.source_features
            ]
        else:
            return [
                DataSource(identifier, source_feature)
                for source_feature in self.source_features
                for identifier in measurement_ids
            ]


def get_calculated_data(
    aggregation_type: AggregationType, data: SeriesData
) -> list[CalculatedItem]:
    # TODO: if aggregation type is Replicate(Average), check for required columns
    # Raise if column(s) do not exist
    return [
        CalculatedItem(
            random_uuid_str(),
            reference.name,
            data[float, reference.column],
            reference.unit,
            reference.source,
            reference.source_features,
            reference.column_key,
        )
        for reference in CALCULATED_DATA_REFERENCE.get(aggregation_type, [])
        if data.get(float, reference.column) is not None
    ]


@dataclass
class Group:
    well_identifier: str | None
    group_identifier: str
    target_identifier: str
    calculated_data: list[CalculatedItem]
    calculated_data_ids: dict[str, str]

    @property
    def name(self) -> str:
        return self.group_identifier.split("(")[0].strip()

    @property
    def key(self) -> str:
        if not self.target_identifier:
            return self.group_identifier
        return f"{self.name}_{self.target_identifier}"

    def get_calculated_data_ids(self) -> dict[str, str]:
        return {
            calculated_data.name: calculated_data.identifier
            for calculated_data in self.calculated_data
        }

    @staticmethod
    def create(data: SeriesData) -> Group:
        well_identifier = data.get(str, "Well")
        calculated_data_items = (
            get_calculated_data(AGGREGATION_LOOKUP[well_identifier], data)
            if well_identifier in AGGREGATION_LOOKUP
            else []
        )
        calculated_data_ids = {
            calculated_data.name: calculated_data.identifier
            for calculated_data in calculated_data_items
        }
        return Group(
            well_identifier=well_identifier,
            group_identifier=data[str, "Group"],
            target_identifier=data[str, "Target"],
            calculated_data=calculated_data_items,
            calculated_data_ids=calculated_data_ids,
        )

    @staticmethod
    def create_rows(data: pd.DataFrame) -> list[Group]:
        # If there is no Group column, we expect there to be no calculated data.
        if "Group" not in data:
            return []
        data = data[data["Sample"].isna()]
        return map_rows(data, Group.create)


@dataclass
class WellItem:
    name: str
    measurement_identifier: str
    well_identifier: str
    group_identifier: str
    target_identifier: str
    run_identifier: str
    instrument_identifier: str
    timestamp: str
    total_partition_count: float
    reporter_dye_setting: str
    concentration: float
    positive_partition_count: float
    plate_identifier: str | None = None
    negative_partition_count: float | None = None
    confidence_interval__95__: float | None = None
    passive_reference_dye_setting: str | None = None
    flourescence_intensity_threshold_setting: float | None = None
    reporter_dye_data_cube: DataCube | None = None
    passive_reference_dye_data_cube: DataCube | None = None
    errors: list[Error] | None = None
    calculated_data: list[CalculatedDataItem] | None = None
    extra_data: dict[str, Any] | None = None

    @property
    def group_key(self) -> str:
        return f"{self.group_identifier}_{self.target_identifier}"

    @staticmethod
    def create(data: SeriesData) -> WellItem:
        measurement_identifier = random_uuid_str()
        return WellItem(
            name=data[str, "Sample"],
            measurement_identifier=measurement_identifier,
            well_identifier=WellItem.get_well_id(data),
            plate_identifier=data.get(str, "Plate"),
            # Group column may be missing if there is no calculated data to group with.
            group_identifier=data.get(str, "Group", ""),
            target_identifier=data.get(str, "Target", NOT_APPLICABLE),
            run_identifier=data[str, "Run"],
            instrument_identifier=data.get(str, "Instrument", NOT_APPLICABLE),
            timestamp=data.get(str, "Date", DEFAULT_EPOCH_TIMESTAMP),
            total_partition_count=round(data[float, "Total"]),
            reporter_dye_setting=data[str, ("Dye", "Channels")],
            concentration=data[float, CONCENTRATION_COLUMNS],
            positive_partition_count=round(data[float, ("Positives", "Count")]),
            confidence_interval__95__=data.get(float, "95%CI"),
            flourescence_intensity_threshold_setting=data.get(float, "Threshold"),
            calculated_data=[
                CalculatedDataItem(
                    identifier=calc_data.identifier,
                    name=calc_data.name,
                    value=calc_data.value,
                    unit=calc_data.unit,
                    data_sources=calc_data.get_data_sources(
                        [measurement_identifier], {}
                    ),
                )
                for calc_data in get_calculated_data(AggregationType.INDIVIDUAL, data)
            ],
            extra_data=data.get_unread(),
        )

    @staticmethod
    def create_cube_well_items(well_data: pd.DataFrame) -> list[WellItem]:
        dye_settings = get_dye_settings(list(well_data.columns))

        # Order rows by data cube index (this is probably always true, but to be safe).
        well_data = well_data.sort_values("Index", axis="index")
        total_partition_count = float(well_data["Index"].max())

        # Assume that any dye setting that does not have a '_target' column is the passive dye setting.
        passive_dye_settings = [
            ds for ds in dye_settings if f"{ds}_target" not in well_data
        ]
        # Assume there is only one passive dye setting.
        if len(passive_dye_settings) != 1:
            msg = f"Expected exactly one possible passive dye setting, got: '{passive_dye_settings}'"
            raise AllotropeConversionError(msg)

        cycle_count = DataCubeComponent(
            FieldComponentDatatype.integer, "cycle count", "#"
        )
        index_column = well_data["Index"].astype(float).tolist()

        data = SeriesData(well_data.iloc[0])
        well_items: list[WellItem] = []
        for dye_setting in [ds for ds in dye_settings if f"{ds}_target" in well_data]:
            positives_column = f"{dye_setting}_pos"
            if positives_column not in well_data:
                msg = f"Expected {positives_column} column to exist."
                raise AllotropeConversionError(msg)
            pos_counts = well_data[positives_column].value_counts().to_dict()
            positive_partition_count = float(pos_counts.get(True, 0))
            negative_partition_count = float(pos_counts.get(False, 0))
            if (
                positive_partition_count + negative_partition_count
            ) != total_partition_count:
                msg = f"positive partition count ({positive_partition_count}) + negative partition count ({negative_partition_count}) != total partition count ({total_partition_count}). This probaby means the values in '{positives_column}' are not bools as expected."

            concentration = data.get(float, CONCENTRATION_COLUMNS)
            errors = []
            if concentration is None:
                concentration = NEGATIVE_ZERO
                errors.append(Error(error="N/A", error_feature="number concentration"))

            well_items.append(
                WellItem(
                    name=data[str, "Sample"],
                    measurement_identifier=random_uuid_str(),
                    well_identifier=data[str, "Well"],
                    plate_identifier=data.get(str, "Plate"),
                    # Group column may be missing if there is no calculated data to group with.
                    group_identifier=data.get(str, "Group", ""),
                    target_identifier=data[str, f"{dye_setting}_target"],
                    run_identifier=data[str, "Run"],
                    instrument_identifier=data.get(str, "Instrument", NOT_APPLICABLE),
                    timestamp=data.get(str, "Date", DEFAULT_EPOCH_TIMESTAMP),
                    total_partition_count=total_partition_count,
                    reporter_dye_setting=dye_setting,
                    passive_reference_dye_setting=passive_dye_settings[0],
                    flourescence_intensity_threshold_setting=data[
                        float, f"{dye_setting}_threshold"
                    ],
                    concentration=concentration,
                    positive_partition_count=positive_partition_count,
                    negative_partition_count=negative_partition_count,
                    confidence_interval__95__=data.get(float, "95%CI"),
                    reporter_dye_data_cube=DataCube(
                        label="reporter dye",
                        structure_dimensions=[cycle_count],
                        structure_measures=[
                            DataCubeComponent(
                                FieldComponentDatatype.double,
                                "reporter dye fluorescence",
                                "RFU",
                            )
                        ],
                        dimensions=[index_column],
                        measures=[well_data[dye_setting].astype(float).tolist()],
                    ),
                    passive_reference_dye_data_cube=DataCube(
                        label="passive reference dye",
                        structure_dimensions=[cycle_count],
                        structure_measures=[
                            DataCubeComponent(
                                FieldComponentDatatype.double,
                                "passive reference dye fluorescence",
                                "RFU",
                            )
                        ],
                        dimensions=[index_column],
                        measures=[
                            well_data[passive_dye_settings[0]].astype(float).tolist()
                        ],
                    ),
                    errors=errors,
                )
            )

        return well_items

    @staticmethod
    def get_dye_settings(columns: list[str]) -> list[str]:
        return [
            col
            for col in columns
            if len(col) in POSSIBLE_DYE_SETTING_LENGTHS and col == col.upper()
        ]

    @staticmethod
    def get_well_id(data: SeriesData) -> str:
        if well_id := data.get(str, "Well", ""):
            return well_id

        sample_name = data[str, "Sample"]
        if match := re.match(r"Sample ([a-zA-Z0-9]+)", sample_name):
            return match.groups()[0]
        msg = f"Unable to get well identifier from sample {sample_name}"
        raise AllotropeConversionError(msg)


@dataclass
class Well:
    items: list[WellItem]

    @staticmethod
    def create_wells(data: pd.DataFrame) -> list[Well]:
        if "Channels" in data:
            subset_col = "Channels"
            groupby_col = "Sample"
        else:
            subset_col = "Sample"
            groupby_col = "Well"
        data = data.dropna(subset=[subset_col])

        well_groups = [well_data for _, well_data in data.groupby(groupby_col)]
        if "Index" in data:
            return [
                Well(WellItem.create_cube_well_items(well_data))
                for well_data in well_groups
            ]
        else:
            return [
                Well(map_rows(well_data, WellItem.create)) for well_data in well_groups
            ]


def create_measurement_groups(wells: list[Well]) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(
            experimental_data_identifier=well.items[0].run_identifier,
            plate_well_count=PLATE_WELL_COUNT,
            measurements=[
                Measurement(
                    identifier=item.measurement_identifier,
                    sample_identifier=item.name,
                    location_identifier=item.well_identifier,
                    measurement_time=item.timestamp,
                    plate_identifier=item.plate_identifier,
                    target_identifier=item.target_identifier,
                    total_partition_count=item.total_partition_count,
                    concentration=item.concentration,
                    positive_partition_count=item.positive_partition_count,
                    negative_partition_count=item.negative_partition_count,
                    confidence_interval__95__=item.confidence_interval__95__,
                    reporter_dye_setting=item.reporter_dye_setting,
                    passive_reference_dye_setting=item.passive_reference_dye_setting,
                    flourescence_intensity_threshold_setting=item.flourescence_intensity_threshold_setting,
                    reporter_dye_data_cube=item.reporter_dye_data_cube,
                    passive_reference_dye_data_cube=item.passive_reference_dye_data_cube,
                    errors=item.errors,
                    custom_info=item.extra_data,
                )
                for item in well.items
            ],
        )
        for well in wells
    ]


def create_calculated_data(
    wells: list[Well], groups: list[Group], common_columns: list[str]
) -> list[CalculatedDataItem]:
    if not groups:
        return []
    # Map measurement ids to group keys and get per-measurement calculated data items
    group_to_ids = defaultdict(list)
    calculated_data_items: list[CalculatedDataItem] = []
    for well in wells:
        for item in well.items:
            group_to_ids[item.group_key].append(item.measurement_identifier)
            group_to_ids[item.group_identifier].append(item.measurement_identifier)
            calculated_data_items.extend(item.calculated_data or [])

    # When parsing a "Summary" file, some calculated columns are across all samples in a group, while others
    # are across a (group, target) pair. common_columns tells us which columns are across the group.
    # For these "common columns" extract them from (group, target) groups into (group,) only groups, so that
    # they are attributed correctly and not duplicated.
    group_id_to_group: dict[str, Group] = {}
    summary_calculated_data: defaultdict[str, dict[str, CalculatedItem]] = defaultdict(
        dict
    )
    for group in groups:
        group_id_to_group[group.group_identifier] = group
        pruned: list[CalculatedItem] = []
        for calculated_data in group.calculated_data:
            if calculated_data.column in common_columns:
                existing_entry = summary_calculated_data.get(
                    group.group_identifier, {}
                ).get(calculated_data.column)
                if existing_entry and existing_entry.value != calculated_data.value:
                    msg = f"Mismatch in value within group for summary calculated data column: '{calculated_data.column}': '{existing_entry.value}' vs '{calculated_data.value}'"
                    raise AllotropeConversionError(msg)
                summary_calculated_data[group.group_identifier][
                    calculated_data.column
                ] = calculated_data
            else:
                pruned.append(calculated_data)
        group.calculated_data = pruned

    for group_id, calc_column_to_calc_data in summary_calculated_data.items():
        groups.append(
            Group(
                well_identifier=group_id_to_group[group_id].well_identifier,
                group_identifier=group_id,
                target_identifier="",
                calculated_data=list(calc_column_to_calc_data.values()),
                calculated_data_ids=group_id_to_group[group_id].calculated_data_ids,
            )
        )

    calculated_data_items.extend(
        [
            CalculatedDataItem(
                identifier=calculated_data.identifier,
                name=calculated_data.name,
                value=calculated_data.value,
                unit=calculated_data.unit,
                data_sources=[
                    DataSource(source.identifier, source.feature)
                    for source in calculated_data.get_data_sources(
                        group_to_ids[group.key], group.calculated_data_ids
                    )
                ],
            )
            for group in groups
            for calculated_data in group.calculated_data
        ]
    )
    return calculated_data_items


def create_metadata(device_identifier: str, file_path: str) -> Metadata:
    return Metadata(
        device_identifier=device_identifier,
        brand_name=BRAND_NAME,
        device_type=DEVICE_TYPE,
        container_type=ContainerType.well_plate,
        software_name=SOFTWARE_NAME,
        product_manufacturer=PRODUCT_MANUFACTURER,
        file_name=Path(file_path).name,
        unc_path=file_path,
    )



from __future__ import annotations

from pathlib import Path

from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    ContainerType,
)
from allotropy.allotrope.models.shared.definitions.definitions import NaN
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import get_key_or_error
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.thermo_fisher_qubit4 import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str

EMISSION_WAVELENGTH_TO_MEASUREMENT_COLUMN = {
    "green": "Green RFU",
    "far red": "Far Red RFU",
}


def create_measurement_group(data: SeriesData) -> MeasurementGroup:
    std_1_rfu = data.get(float, "Std 1 RFU", validate=SeriesData.NOT_NAN)
    std_2_rfu = data.get(float, "Std 2 RFU", validate=SeriesData.NOT_NAN)
    std_3_rfu = data.get(float, "Std 3 RFU", validate=SeriesData.NOT_NAN)
    return MeasurementGroup(
        measurement_time=data[str, "Test Date"],
        experiment_type=data.get(str, "Assay Name"),
        measurements=[
            Measurement(
                type_=MeasurementType.FLUORESCENCE,
                identifier=random_uuid_str(),
                fluorescence=data[
                    float,
                    get_key_or_error(
                        "wavelength",
                        data.get(str, "Emission", "").lower(),
                        EMISSION_WAVELENGTH_TO_MEASUREMENT_COLUMN,
                    ),
                ],
                batch_identifier=data.get(str, "Run ID"),
                sample_identifier=data[str, "Test Name"],
                sample_volume_setting=data.get(
                    float, "Sample Volume (μL)", validate=SeriesData.NOT_NAN
                ),
                excitation_wavelength_setting=data.get(str, "Excitation"),
                emission_wavelength_setting=data[str, "Emission"],
                dilution_factor_setting=data.get(
                    float, "Dilution Factor", validate=SeriesData.NOT_NAN
                ),
                original_sample_concentration=data.get(
                    float, "Original sample conc.", NaN
                ),
                original_sample_concentration_unit=data.get(
                    str, "Units_Original sample conc."
                ),
                sample_custom_info={
                    "qubit tube concentration": {
                        "value": data.get(float, "Qubit® tube conc.", NaN),
                        "unit": data.get(str, "Units_Qubit® tube conc.", UNITLESS),
                    },
                    "standard 1 concentration": {"value": std_1_rfu, "unit": "RFU"}
                    if std_1_rfu is not None
                    else None,
                    "standard 2 concentration": {
                        "value": std_2_rfu,
                        "unit": "RFU",
                    }
                    if std_2_rfu is not None
                    else None,
                    "standard 3 concentration": {
                        "value": std_3_rfu,
                        "unit": "RFU",
                    }
                    if std_3_rfu is not None
                    else None,
                },
            )
        ],
    )


def create_metadata(file_path: str) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        device_identifier=NOT_APPLICABLE,
        model_number=constants.MODEL_NUMBER,
        software_name=constants.QUBIT_SOFTWARE,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        brand_name=constants.BRAND_NAME,
        device_type=constants.DEVICE_TYPE,
        container_type=ContainerType.tube,
    )



from __future__ import annotations

from allotropy.allotrope.models.adm.cell_counting.rec._2024._09.cell_counting import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Data,
    Mapper,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_reader import create_reader_data
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_structure import (
    create_measurement_group,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class ViCellXRParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Beckman Vi-Cell XR"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "txt,xls,xlsx"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader_data = create_reader_data(named_file_contents)
        if not reader_data.data:
            msg = "Cannot parse ASM from empty file."
            raise AllotropeConversionError(msg)

        return Data(
            create_metadata(reader_data, named_file_contents.original_file_path),
            [create_measurement_group(row) for row in reader_data.data],
        )



from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    Measurement,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCubeComponent
from allotropy.parsers.cytiva_unicorn.constants import DEVICE_TYPE
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.transformations import (
    MScm2Sm,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.generic import (
    UnicornMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.static_docs import (
    StaticDocs,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)


class ConductivityMeasurement(UnicornMeasurement):
    @classmethod
    def create(
        cls,
        handler: UnicornZipHandler,
        elements: list[StrictXmlElement],
        static_docs: StaticDocs,
    ) -> Measurement:
        return cls.get_measurement(
            static_docs=static_docs,
            chromatogram_data_cube=cls.get_data_cube(
                handler,
                cls.filter_curve(elements, r"^Cond$"),
                DataCubeComponent(
                    type_=FieldComponentDatatype.float,
                    concept="electric conductivity",
                    unit="S/m",
                ),
                transformation=MScm2Sm(),
            ),
            processed_data_chromatogram_data_cube=cls.get_data_cube(
                handler,
                cls.filter_curve(elements, r"^% Cond$"),
                DataCubeComponent(
                    type_=FieldComponentDatatype.float,
                    concept="electric conductivity",
                    unit="%",
                ),
            ),
            device_control_docs=[
                DeviceControlDoc(
                    device_type=DEVICE_TYPE,
                )
            ],
        )



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, EMPTY_CSV_LINE, read_to_lines


class ExampleWeylandYutaniReader:
    SUPPORTED_EXTENSIONS = "csv"

    middle: pd.DataFrame | None
    bottom: pd.DataFrame | None

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader(read_to_lines(named_file_contents))
        _ = reader.pop_csv_block_as_df(empty_pat=EMPTY_CSV_LINE)
        self.middle = reader.pop_csv_block_as_df(empty_pat=EMPTY_CSV_LINE)
        self.bottom = reader.pop_csv_block_as_df(empty_pat=EMPTY_CSV_LINE)



from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Generic, TypeVar

from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_NAME
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.timestamp_parser import TimestampParser

Data = TypeVar("Data")
Model = TypeVar("Model")
Mapper = TypeVar("Mapper")


class VendorParser(ABC, Generic[Data, Model]):
    """Base class for all vendor parsers."""

    # The display name of the parser. Displayed in the README.
    DISPLAY_NAME: str
    # Signifies if the parser is ready to be used. Can be set to ReleaseState.WORKING_DRAFT while being developed.
    RELEASE_STATE: ReleaseState
    # Comma separated list of file extensions that this parser supports.
    SUPPORTED_EXTENSIONS: str
    # The schema mapper to use for mapping to ASM
    SCHEMA_MAPPER: Callable[..., SchemaMapper[Data, Model]]

    timestamp_parser: TimestampParser

    def __init__(self, timestamp_parser: TimestampParser | None = None):
        self.timestamp_parser = timestamp_parser or TimestampParser()

    def _get_mapper(self) -> SchemaMapper[Data, Model]:
        return self.SCHEMA_MAPPER(self.asm_converter_name, self._get_date_time)

    @abstractmethod
    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        raise NotImplementedError

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        return self._get_mapper().map_model(self.create_data(named_file_contents))

    @property
    def asm_converter_name(self) -> str:
        return f'{ASM_CONVERTER_NAME}_{self.DISPLAY_NAME.replace(" ", "_").replace("-", "_")}'.lower()

    def _get_date_time(self, time: str) -> TDateTimeValue:
        return self.timestamp_parser.parse(time)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import NaN
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    CalculatedDataItem,
    Data,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
    ProcessedDataFeature,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float


@dataclass
class MetadataNanodrop:
    nucleic_acid_unit: str
    absorbance_columns: list[str]
    calculated_columns: list[str]


@dataclass
class InputNanodrop:
    metadata: MetadataNanodrop
    data: list[SeriesData]


@dataclass(frozen=True)
class MeasurementGroupNanodrop(MeasurementGroup):
    @staticmethod
    def create(
        row: SeriesData, metadata: MetadataNanodrop, experiment_type: str
    ) -> MeasurementGroupNanodrop:
        sample_identifier = row[(str, "Sample Name")]
        nucleic_acid = row.get(float, "Nucleic Acid", validate=SeriesData.NOT_NAN)
        baseline_absorbance = row.get(
            float, "Baseline Absorbance", validate=SeriesData.NOT_NAN
        )
        electronic_absorbance_reference_wavelength_setting = row.get(
            float, "Baseline Correction", validate=SeriesData.NOT_NAN
        )
        nucleic_acid_factor = row.get(
            float, "Nucleic Acid Factor", validate=SeriesData.NOT_NAN
        )

        measurements = [
            Measurement(
                type_=MeasurementType.ULTRAVIOLET_ABSORBANCE,
                identifier=random_uuid_str(),
                sample_identifier=sample_identifier,
                absorbance=row.get(float, column, NaN),
                baseline_absorbance=baseline_absorbance,
                electronic_absorbance_reference_wavelength_setting=electronic_absorbance_reference_wavelength_setting,
                detector_wavelength_setting=try_float(
                    column.removeprefix("A"),
                    "detector wavelength setting",
                ),
                processed_data=(
                    None
                    if nucleic_acid is None
                    else ProcessedData(
                        features=[
                            ProcessedDataFeature(
                                result=nucleic_acid,
                                unit=metadata.nucleic_acid_unit,
                            )
                        ],
                    )
                ),
                device_control_custom_info={
                    "nucleic acid factor": {
                        "value": nucleic_acid_factor,
                        "unit": UNITLESS,
                    }
                    if nucleic_acid_factor is not None
                    else None
                },
            )
            for column in metadata.absorbance_columns
        ]

        return MeasurementGroupNanodrop(
            measurements=measurements,
            measurement_time=row[str, "Date"],
            experiment_type=experiment_type,
        )

    def get_measurement(self, raw_wavelength: str) -> Measurement | None:
        detector_wavelength_setting = try_float(
            raw_wavelength.removeprefix("A"),
            "detector wavelength setting",
        )
        for measurement in self.measurements:
            if measurement.detector_wavelength_setting == detector_wavelength_setting:
                return measurement
        return None


def create_calculated_data(
    row: SeriesData,
    metadata: MetadataNanodrop,
    measurement_group: MeasurementGroupNanodrop,
) -> list[CalculatedDataItem]:
    calculated_data = []
    for column in metadata.calculated_columns:
        value = row.get(float, column)
        if value is None:
            continue

        calculated_data.append(
            CalculatedDataItem(
                identifier=random_uuid_str(),
                name=column,
                value=value,
                unit=UNITLESS,
                data_sources=[
                    DataSource(
                        identifier=measurement.identifier,
                        feature="absorbance",
                    )
                    for wavelength in column.split("/")
                    if (measurement := measurement_group.get_measurement(wavelength))
                ],
            )
        )
    return calculated_data


@dataclass
class RowElements:
    # is not a list to avoid invariance problem
    measurement_groups: tuple[MeasurementGroup, ...]
    calculated_data: list[CalculatedDataItem]

    @staticmethod
    def filter_column_names(data: pd.DataFrame, regex: str) -> list[str]:
        return [column for column in data.columns if re.search(regex, column)]

    @staticmethod
    def filter_column_name(data: pd.DataFrame, pattern: str) -> str:
        columns = RowElements.filter_column_names(data, pattern)
        if len(columns) == 1:
            return columns[0]

        error = f"Unable to find column using pattern {pattern}."
        raise AllotropeConversionError(error)

    @staticmethod
    def filter_data(data: pd.DataFrame) -> InputNanodrop:
        nucleic_acid_col = RowElements.filter_column_name(data, r"^Nucleic Acid\s*\(")
        absorbance_columns = RowElements.filter_column_names(data, r"^A(\d{3})$")
        calculated_columns = RowElements.filter_column_names(
            data, r"^A(\d{3})/A(\d{3})$"
        )
        baseline_correction_col = RowElements.filter_column_name(
            data, "^Baseline Correction"
        )

        columns = [
            "Date",
            "Sample Name",
            "Baseline Absorbance",
            "Nucleic Acid Factor",
            baseline_correction_col,
            nucleic_acid_col,
            *absorbance_columns,
            *calculated_columns,
        ]
        new_data = data[columns].rename(
            columns={
                baseline_correction_col: "Baseline Correction",
                nucleic_acid_col: "Nucleic Acid",
            }
        )

        if match := re.match(r".*\((.+)\)", nucleic_acid_col):
            nucleic_acid_unit = match.group(1)
        else:
            error = "Unable to find nucleic acid unit"
            raise AllotropeConversionError(error)

        return InputNanodrop(
            metadata=MetadataNanodrop(
                nucleic_acid_unit,
                absorbance_columns,
                calculated_columns,
            ),
            data=[SeriesData(row) for _, row in new_data.iterrows()],
        )

    @staticmethod
    def create(data: pd.DataFrame, experiment_type: str) -> RowElements:
        clean_data = RowElements.filter_data(data)

        measurement_groups = []
        calculated_data = []
        for row in clean_data.data:
            measurement_group = MeasurementGroupNanodrop.create(
                row, clean_data.metadata, experiment_type
            )
            measurement_groups.append(measurement_group)
            calculated_data += create_calculated_data(
                row, clean_data.metadata, measurement_group
            )

        return RowElements(tuple(measurement_groups), calculated_data)


@dataclass(frozen=True)
class DataNanodrop(Data):
    @staticmethod
    def create(data: pd.DataFrame, experiment_type: str, file_path: str) -> Data:
        row_elements = RowElements.create(data, experiment_type)

        return Data(
            metadata=Metadata(
                device_identifier="N/A",
                device_type="absorbance detector",
                model_number="NanoDrop One",
                brand_name="NanoDrop",
                product_manufacturer="ThermoFisher Scientific",
                file_name=Path(file_path).name,
                unc_path=file_path,
                software_name="NanoDrop One software",
            ),
            measurement_groups=list(row_elements.measurement_groups),
            calculated_data=row_elements.calculated_data or None,
        )



from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ContainerType

DEVICE_TYPE = "qPCR"
SOFTWARE_NAME = "Thermo QuantStudio"
SOFTWARE_VERSION = "1.0"
DATA_SYSTEM_INSTANCE_IDENTIFIER = "localhost"
CONTAINER_TYPE = ContainerType.qPCR_reaction_block

POSSIBLE_WELL_COUNTS = [1, 2, 4, 6, 8, 12, 24, 48, 72, 96, 384, 1536, 3456]



from collections import defaultdict
from dataclasses import dataclass
import json
from pathlib import Path
import re
from typing import Any

from allotropy.allotrope.schema_parser.backup_manager import is_backup_file
from allotropy.allotrope.schema_parser.path_util import SHARED_SCHEMAS_PATH


# Used to match and locate the model definiton defined in allotropy/allotrope/models/shared
@dataclass
class SchemaModel:
    name: str
    # The ASM schema of the model, used to match against schemas when refactoring genrated models.
    schema: dict[str, Any]
    # The location of the defined shared class, relative to allotropy.allotrope.models.shared
    import_info: tuple[str, str]


# TODO: roughly copied from datamodel-codegen. There is probably more to use/import here to cover all cases.
def snake_to_upper_camel(word: str, delimiter: str = "_") -> str:
    prefix = ""
    if word.startswith(delimiter):
        prefix = "_"
        word = word[1:]

    return prefix + "".join(
        x[0].upper() + x[1:] for x in re.split(delimiter, word) if x
    )


# TODO: this isn't complete. It covers the current schemas in shared/models, but needs:
# - Hardening (e.g. check type and use that to find properties)
# - More cases (e.g. oneOf / allOf)
def get_all_schema_components(schema: dict[str, Any]) -> dict[str, dict[str, Any]]:
    if "items" in schema:
        schema = schema["items"]

    if "allOf" in schema:
        mapping = {}
        for item in schema["allOf"]:
            mapping.update(get_all_schema_components(item))
        return mapping

    if not isinstance(schema, dict) or "properties" not in schema:
        return {}

    mapping = {}
    for prop_name, prop in schema["properties"].items():
        if "oneOf" in prop:
            count = 0
            for item in prop["oneOf"]:
                if item.get("type") == "object":
                    mapping[f"{prop_name} item{count if count else ''}"] = item
                    mapping.update(get_all_schema_components(item))
                    count += 1
        else:
            mapping[prop_name] = prop
            mapping.update(get_all_schema_components(prop))

    return mapping


def get_schema_definitions_mapping() -> dict[str, list[SchemaModel]]:
    schema_mapping = defaultdict(list)

    definition_files = [
        (path.name, directory)
        for directory in ["definitions", "components"]
        for path in Path(SHARED_SCHEMAS_PATH, directory).iterdir()
        if not is_backup_file(path)
    ]

    for schema_file, directory in definition_files:
        schema_module = f"{directory}.{schema_file[:-5]}"

        with open(Path(SHARED_SCHEMAS_PATH, directory, schema_file)) as f:
            schemas = json.load(f)

        for schema_name, schema in schemas.items():
            schema_mapping[schema_name].append(
                SchemaModel(
                    schema_name,
                    schema,
                    (schema_module, snake_to_upper_camel(schema_name, "\\W")),
                )
            )
            components = get_all_schema_components(schema)
            for component_name, component_schema in components.items():
                schema_mapping[component_name].append(
                    SchemaModel(
                        component_name,
                        component_schema,
                        (schema_module, snake_to_upper_camel(component_name, "\\W")),
                    )
                )

    return dict(schema_mapping)



from __future__ import annotations

from xml.etree import ElementTree

# xml fromstring is vulnerable so defusedxml version is used instead
from defusedxml.ElementTree import fromstring

from allotropy.parsers.utils.values import assert_not_none, try_float


class StrictXmlElement:
    @classmethod
    def create_from_bytes(cls, data: bytes) -> StrictXmlElement:
        return StrictXmlElement(fromstring(data))

    def __init__(self, element: ElementTree.Element):
        self.element = element

    def find(self, name: str) -> StrictXmlElement:
        return StrictXmlElement(
            assert_not_none(
                self.element.find(name),
                msg=f"Unable to find '{name}' in xml file contents",
            )
        )

    def recursive_find(self, names: list[str]) -> StrictXmlElement:
        if len(names) == 0:
            return self
        name, *sub_names = names
        return self.find(name).recursive_find(sub_names)

    def findall(self, name: str) -> list[StrictXmlElement]:
        return [StrictXmlElement(element) for element in self.element.findall(name)]

    def get_attr(self, name: str) -> str:
        return str(
            assert_not_none(
                self.element.get(name),
                msg=f"Unable to find '{name}' in xml file contents",
            )
        )

    def get_text(self) -> str:
        return str(self.element.text)

    def get_float(self, name: str) -> float:
        return try_float(self.get_text(), name)



from __future__ import annotations

from pathlib import Path
from re import search

from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    Metadata,
)
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)


def get_audit_trail_entry(element: StrictXmlElement) -> StrictXmlElement | None:
    audit_trail_entries = element.recursive_find(["AuditTrail", "AuditTrailEntries"])
    for element in audit_trail_entries.findall("AuditTrailEntry"):
        if element.find("GroupName").get_text() == "EvaluationLoggingStarted":
            return element
    return None


def get_audit_trail_entry_user(handler: UnicornZipHandler) -> str:
    evaluation_log = handler.get_evaluation_log()
    if audit_trail_entry := get_audit_trail_entry(evaluation_log):
        if match := search(
            r"User: (.+)\. ",
            audit_trail_entry.find("LogEntry").get_text(),
        ):
            return match.group(1)
    return "Default"


def create_metadata(
    handler: UnicornZipHandler, results: StrictXmlElement, file_path: str
) -> Metadata:
    system_data = handler.get_system_data()
    instrument_config_data = handler.get_instrument_config_data()

    instrument_config = system_data.recursive_find(
        ["System", "InstrumentConfiguration"]
    )

    return Metadata(
        asset_management_identifier=instrument_config.get_attr("Description"),
        product_manufacturer="Cytiva Life Sciences",
        device_identifier=results.find("SystemName").get_text(),
        firmware_version=instrument_config_data.find("FirmwareVersion").get_text(),
        analyst=get_audit_trail_entry_user(handler),
        file_name=Path(file_path).name,
        unc_path=file_path,
    )



from enum import Enum


class ReleaseState(Enum):
    # Ready for production use
    RECOMMENDED = "RECOMMENDED"
    # Working for some cases, may have bugs or need more test cases for hardening
    CANDIDATE_RELEASE = "CANDIDATE_RELEASE"
    # In development, not recommended for production use
    WORKING_DRAFT = "WORKING_DRAFT"

    def __str__(self) -> str:
        return self.value.replace("_", " ").title()



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ScanPositionSettingPlateReader,
    TransmittedLightSetting,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType

SOFTWARE_NAME = "Kaleido"
DEVICE_IDENTIFIER = "EnSight"
MODEL_NUMBER = "EnSight"
PRODUCT_MANUFACTURER = "Revvity"
MAGNIFICATION_SETTING = 4

CHANNEL_COLUMNS_ERROR = "Expected every Channel be followed by: Excitation wavelength [nm], Excitation Power [%], Exposure Time [ms], Additional Focus offset [mm]"


PLATEMAP_TO_SAMPLE_ROLE_TYPE = {
    "B": SampleRoleType.blank_role,
    "C": SampleRoleType.control_sample_role,
    "S": SampleRoleType.standard_sample_role,
    "U": SampleRoleType.unknown_sample_role,
    "E": SampleRoleType.control_sample_role,
    "ZL": SampleRoleType.control_sample_role,
    "ZH": SampleRoleType.control_sample_role,
    "LB": SampleRoleType.control_sample_role,
    "LC": SampleRoleType.control_sample_role,
    "LH": SampleRoleType.control_sample_role,
}


SCAN_POSITION_CONVERSION = {
    "TOP": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
    "BOTTOM": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
}


TRANSMITTED_LIGHT_CONVERSION = {
    "BRIGHTFIELD": TransmittedLightSetting.brightfield,
    "DARKFIELD": TransmittedLightSetting.darkfield,
    "PHASE CONTRAST": TransmittedLightSetting.phase_contrast,
}



from __future__ import annotations

from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import (
    Measurement,
    Metadata,
)
from allotropy.exceptions import get_key_or_error
from allotropy.parsers.constants import DEFAULT_EPOCH_TIMESTAMP
from allotropy.parsers.qiacuity_dpcr.constants import (
    BRAND_NAME,
    DEVICE_IDENTIFIER,
    DEVICE_TYPE,
    PRODUCT_MANUFACTURER,
    SAMPLE_ROLE_TYPE_MAPPING,
    SOFTWARE_NAME,
)
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_measurements(data: SeriesData) -> Measurement:
    sample_role_type = data.get(str, "Type")
    # TODO: When the sample role type model is updated in this repo, we should update this
    # Map sample role types to valid sample role types from ASM
    if sample_role_type is not None:
        sample_role_type = get_key_or_error(
            "sample type", sample_role_type, SAMPLE_ROLE_TYPE_MAPPING
        )

    return Measurement(
        identifier=random_uuid_str(),
        measurement_time=DEFAULT_EPOCH_TIMESTAMP,
        sample_identifier=data[str, "Sample/NTC/Control"],
        sample_role_type=sample_role_type,
        location_identifier=data[str, "Well Name"],
        plate_identifier=data.get(str, "Plate ID"),
        target_identifier=data[str, "Target"],
        total_partition_count=data[int, "Partitions (valid)"],
        concentration=data[float, ["Concentration (copies/μL)", "Conc. [copies/μL]"]],
        positive_partition_count=data[int, "Partitions (positive)"],
        negative_partition_count=data.get(int, "Partitions (negative)"),
        flourescence_intensity_threshold_setting=data.get(float, "Threshold"),
    )


def create_metadata(file_path: str) -> Metadata:
    return Metadata(
        device_identifier=DEVICE_IDENTIFIER,
        brand_name=BRAND_NAME,
        device_type=DEVICE_TYPE,
        software_name=SOFTWARE_NAME,
        product_manufacturer=PRODUCT_MANUFACTURER,
        file_name=Path(file_path).name,
        unc_path=file_path,
    )



from typing import ClassVar

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ExperimentType
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_calculated_documents import (
    iter_standard_curve_calc_docs,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_views import (
    SampleView,
    TargetRoleView,
    TargetView,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.creator import (
    Creator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Data,
    Header,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.standard_curve.structure import (
    StandardCurveWellList,
)


class StandardCurveCreator(Creator):
    PLUGIN_REGEX: ClassVar[str] = r"Standard Curve"
    EXPECTED_SHEETS: ClassVar[list[str]] = [
        "Standard Curve Result",
    ]

    @classmethod
    def create(cls, reader: DesignQuantstudioReader) -> Data:
        header = Header.create(reader.header)
        wells = StandardCurveWellList.create(reader, header)
        well_items = wells.get_well_items()

        return Data(
            header,
            wells,
            experiment_type=ExperimentType.standard_curve_qPCR_experiment,
            calculated_documents=list(
                iter_standard_curve_calc_docs(
                    view_st_data=SampleView(sub_view=TargetView()).apply(well_items),
                    view_tr_data=TargetRoleView().apply(well_items),
                )
            ),
            reference_target=None,
            reference_sample=None,
        )
